"""Top level module for uetools"""

__descr__ = "Unreal Engine Tools"
__version__ = "0.2.3"
__license__ = "BSD 3-Clause License"
__author__ = "Pierre Delaunay"
__author_email__ = "pierre@delaunay.io"
__copyright__ = "2022 Pierre Delaunay"
__url__ = "https://github.com/kiwi-lang/uetools"


from .cli import args, main

__slot___ = ["main", "args"]



import os
import re
from dataclasses import dataclass
from typing import Optional

from uetools.core.command import Command, newparser
from uetools.core.conf import CONFIG, CONFIGNAME, get_version_tag, load_conf, save_conf


@dataclass
class Arguments:
    """Initialize the configuration file with unreal engine folders

    Attributes
    ----------
    engine: str
        Path to the unreal engine folder (C:/opt/UnrealEngine/Engine)

    project: str
        Path to the unreal project folder (C:/Projects)

    version: str
        Engine version

    Examples
    --------

    .. code-block:: console

       uecli init --engine C:/opt/UnrealEngine/Engine --projects C:/opt/Projects

    """

    # fmt: off
    engine  : Optional[str] = None  # Path to the unreal engine folder (C:/opt/UnrealEngine/Engine)
    projects: Optional[str] = None  # Path to the unreal project folder (C:/Projects)
    version : Optional[str] = None  # Unreal Engine Version (5.1)
    # fmt: on


BUILTIN_PATTERN = re.compile(r"UE_(?P<version>(([0-9]*)\.[0-9]*))")


def get_engine_version(path):
    # Prebuild installs have a path that contains UEM.MIN/
    result = BUILTIN_PATTERN.search(path)
    if result:
        version = result.groupdict().get("version")
        if version:
            return version

    return get_version_tag(path, "archive")


class Init(Command):
    """Initialize the configuration file for the command line interface"""

    name: str = "init"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, Init)
        parser.add_arguments(Arguments, dest="init")

    @staticmethod
    def execute(args):
        """Initialize the engine and projects folders"""
        args = args.init

        config = os.path.join(CONFIG, CONFIGNAME)
        conf = {}

        default_engine = "/UnrealEngine/Engine"
        default_project = os.path.abspath(os.path.join("..", default_engine))

        if os.path.exists(config):
            conf = load_conf()
            default_engine = conf.get("engine_path", default_engine)
            default_project = conf.get("project_path", default_project)

        if args.engine is None:
            engine_path = input(f"Engine Folder [{default_engine}]: ")
        else:
            engine_path = args.engine

        if args.projects is None:
            project_folders = input(f"Project Folder [{default_project}]: ")
        else:
            project_folders = args.projects

        engine_path = engine_path or default_engine
        project_folders = project_folders or default_project

        conf["engine_path"] = engine_path
        conf["project_path"] = project_folders

        EngineAdd.addengine(conf, args.version, engine_path)

        save_conf(conf)
        print(f"Updated Engine paths inside `{config}`")
        return 0


@dataclass
class EngineAddArguments:
    """Add an engine version

    Attributes
    ----------
    version: str
        Version of name of the engine version

    engine: str
        Path to the unreal engine folder (C:/opt/UnrealEngine/Engine)

    Examples
    --------

    .. code-block:: console

       uecli engine-add --version src --engine C:/opt/UnrealEngine/Engine

    """

    # fmt: off
    version: Optional[str] = None   # Unreal Engine Version (5.1)
    engine : Optional[str] = None   # Path to the unreal engine folder (C:/opt/UnrealEngine/Engine)
    force  : bool          = False  # Allow version override
    # fmt: on


class EngineAdd(Command):
    """Add an unreal engine version"""

    name: str = "engine-add"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, EngineAdd)
        parser.add_arguments(EngineAddArguments, dest="args")

    @staticmethod
    def execute(args):
        config = os.path.join(CONFIG, CONFIGNAME)
        conf = {}

        if os.path.exists(config):
            conf = load_conf()

        EngineAdd.addengine(conf, args.args.version, args.args.engine, args.args.force)

        save_conf(conf)
        return 0

    @staticmethod
    def addengine(conf, version, path, force=False):
        if version is None:
            version = get_engine_version(path)

        engines = conf.get("engines", dict())

        if version not in engines or force:
            engines[version] = path
        else:
            print(
                f"{version} already exists `{engines[version]}` use --force to override"
            )

        conf["engines"] = engines


COMMANDS = [Init, EngineAdd]



import io
import re
from collections import defaultdict
from dataclasses import dataclass

SECTION_REGEX = re.compile(r"\[(?P<section>.*)\]")
KEY_REGEX = re.compile(r"(?P<op>(\+|-|\.|!))?(?P<key>.*)=(?P<value>.*)")
COMMENT_REGEX = re.compile(r";(?P<comment>.*)")

# pylint: disable=too-few-public-methods
class Entry:
    """Base Entry of the INI AST"""


KEY = ""
ADD = "+"
REM = "-"
ADD_PROPERTY = "."
REM_PROPERTY = "!"


@dataclass
class Section(Entry):
    """Section inside an ini file"""

    title: str

    def to_ini(self):
        """Convert the entry to a ini line"""
        return f"[{self.title}]\n"


@dataclass
class Key(Entry):
    """Key value inside an ini file"""

    op: str
    key: str
    value: str

    def to_ini(self):
        """Convert the entry to a ini line"""
        return f"{self.op}{self.key}={self.value}\n"


@dataclass
class Comment(Entry):
    """Comment inside an ini file"""

    comment: str

    def to_ini(self):
        """Convert the entry to a ini line"""
        return f";{self.comment}\n"


@dataclass
class Newline(Entry):
    """New line inside an ini file"""

    def to_ini(self):
        """Convert the entry to a ini line"""
        return "\n"


def fake_config_file():
    """Testing config"""
    return io.StringIO(
        """
[/Script/Section]
+Weird.Key1=(Name="Q",Command="Foo")
-Weird/Key2=(OldClassName="TP_BlankGameModeBase",NewClassName="GamekitDevGameModeBase")
.Weird,Key3=1
!Weird*Key4=True

[Another.Section]
UniqueKey=123

"""
    )


class UnrealINIParser:
    """Parse a Unreal INI configuration file
    It tries to minimize the number of modification it makes when updating
    the configuration file

    Examples
    --------

    >>> ini = UnrealINIParser(fake_config_file())
    >>> ini.insert('Another.Section', 'UniqueKey', 3)
    'updated'

    >>> file = io.StringIO()
    >>> ini.write(file)
    >>> print(file.getvalue())
    <BLANKLINE>
    [/Script/Section]
    +Weird.Key1=(Name="Q",Command="Foo")
    -Weird/Key2=(OldClassName="TP_BlankGameModeBase",NewClassName="GamekitDevGameModeBase")
    .Weird,Key3=1
    !Weird*Key4=True
    <BLANKLINE>
    [Another.Section]
    UniqueKey=3
    <BLANKLINE>
    <BLANKLINE>

    >>> ini.remove('Another.Section', 'UniqueKey')
    'removed'

    >>> file = io.StringIO()
    >>> ini.write(file)
    >>> print(file.getvalue())
    <BLANKLINE>
    [/Script/Section]
    +Weird.Key1=(Name="Q",Command="Foo")
    -Weird/Key2=(OldClassName="TP_BlankGameModeBase",NewClassName="GamekitDevGameModeBase")
    .Weird,Key3=1
    !Weird*Key4=True
    <BLANKLINE>
    [Another.Section]
    <BLANKLINE>
    <BLANKLINE>

    >>> ini.insert('New.Section', 'UniqueKey1', 123)
    'inserted'

    >>> ini.insert('New.Section', 'UniqueKey2', 123)
    'inserted'

    >>> ini.insert('New.Section', 'UniqueKey2', 4)
    'updated'


    >>> file = io.StringIO()
    >>> ini.write(file)
    >>> print(file.getvalue())
    <BLANKLINE>
    [/Script/Section]
    +Weird.Key1=(Name="Q",Command="Foo")
    -Weird/Key2=(OldClassName="TP_BlankGameModeBase",NewClassName="GamekitDevGameModeBase")
    .Weird,Key3=1
    !Weird*Key4=True
    <BLANKLINE>
    [Another.Section]
    <BLANKLINE>
    [New.Section]
    UniqueKey1=123
    UniqueKey2=4
    <BLANKLINE>

    """

    def __init__(self, fp) -> None:
        # Main references: keeps the file order stable
        # to minimise diffs on updates
        self.sections = [None]

        # Index
        self.entries = defaultdict(list)
        self.entries[None] = []

        # for unique section, key pairs
        self.entries_key = defaultdict(list)

        # Parsing state
        self.current_section = None

        self.parse_file(fp)

    def insert(self, section, key, value, op=KEY):
        """Insert or update a new key"""
        if op == KEY:
            # Change the value of an existing entry
            if (section, key) in self.entries_key:
                entry = self.entries_key[(section, key)][0]
                entry.value = value
                return "updated"

            # add a new entry to the section
            entries = self.entries.get(section, None)

            if entries is None:
                entry = Section(section)
                self.sections.append(entry)

                entries = self.entries[section]

            entry = Key(op, key, value)
            entries.append(entry)

            self.entries_key[(section, key)].append(entry)
            return "inserted"

        # we do not need the functionality right now
        raise NotImplementedError()

    def remove(self, section, key, op=KEY):
        """Remove a key"""
        if op == KEY:
            # Change the value of an existing entry
            if (section, key) in self.entries_key:
                entry = self.entries_key.pop((section, key))[0]

                entries = self.entries[section]
                entries.remove(entry)
                return "removed"

            return "missing"

        # we do not need the functionality right now
        raise NotImplementedError()

    def write(self, fp):
        """Write the current configuration to a file"""
        for section in self.sections:
            if section:
                fp.write(section.to_ini())

            title = section.title if section else None
            entries = self.entries[title]

            for entry in entries:
                fp.write(entry.to_ini())

    def parse_file(self, fp):
        """Parse a file"""
        for line in fp.readlines():
            self.parse_line(line)

    def parse_line(self, line):
        """Parse a line"""
        line = line.strip()

        if line == "":
            self.entries[self.current_section].append(Newline())
            return

        result = KEY_REGEX.search(line)
        if result:
            self.process_key(**result.groupdict())
            return

        result = SECTION_REGEX.search(line)
        if result:
            self.process_section(**result.groupdict())
            return

        result = COMMENT_REGEX.search(line)
        if result:
            self.process_comment(**result.groupdict())
            return

    def process_key(self, op, key, value):
        """Parse a key"""
        assert key is not None
        assert self.current_section is not None
        if op is None:
            op = KEY

        entry = Key(op, key, value)

        self.entries[self.current_section].append(entry)
        self.entries_key[(self.current_section, key)].append(entry)

    def process_section(self, section):
        """Parse a section"""
        entry = Section(section)

        self.sections.append(entry)
        self.current_section = section

    def process_comment(self, comment):
        """Parse a comment"""
        entry = Section(comment)
        self.entries[self.current_section].append(entry)



import os
from dataclasses import dataclass
from io import UnsupportedOperation

from uetools.core.command import Command, newparser
from uetools.core.conf import editor, find_project
from uetools.core.ini import UnrealINIParser


@dataclass
class Arguments:
    """Tweak your project settings to enable python scripting in your project

    Attributes
    ----------
    project: str
        Name of the project

    """

    project: str


class Python(Command):
    """Tweak your project settings to enable python scripting in your project"""

    name: str = "python"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, Python)
        parser.add_argument(
            "project", default=None, type=str, help="name of your project"
        )

    @staticmethod
    def execute(args):
        project = find_project(args.project)
        folder = os.path.dirname(project)

        conf = os.path.join(folder, "Config")
        default_engine = os.path.join(conf, "DefaultEngine.ini")

        with open(default_engine, encoding="utf-8") as file:
            config = UnrealINIParser(file)

        python_section = "/Script/PythonScriptPlugin.PythonScriptPluginUserSettings"
        config.insert(python_section, "bDeveloperMode", "True")
        config.insert(python_section, "bEnableContentBrowserIntegration", "True")

        python_setting = "/Script/PythonScriptPlugin.PythonScriptPluginSettings"
        config.insert(python_setting, "bDeveloperMode", "True")

        if os.path.exists(os.path.join(folder, args.project, "Plugings", "Gamekit")):
            base_path = os.path.join(args.project, "Plugins", "Gamekit", "Script")
            relative_path = os.path.relpath(folder, os.path.dirname(editor()))
            final_path = os.path.join(relative_path, base_path)
            config.insert(python_setting, "+AdditionalPaths", f'(Path="{final_path}")')

        try:
            with open(default_engine, "w", encoding="utf-8") as file:
                config.write(file)

            return 0
        except UnsupportedOperation:
            print("Could not save config")
            print("Is unreal engine open ?")
            raise


COMMANDS = Python



from __future__ import annotations

from dataclasses import asdict, dataclass

from simple_parsing import choice

from uetools.core.command import Command, command_builder, newparser
from uetools.core.conf import find_project, get_build_modes, get_build_platforms, ubt
from uetools.core.run import run

modes = [
    "AggregateParsedTimingInfo",
    "Analyze",
    "Build",
    "Clean",
    "Deploy",
    "Execute",
    "GenerateClangDatabase",
    "GenerateProjectFiles",
    "IOSPostBuildSync",
    "JsonExport",
    "ParseMsvcTimingInfo",
    "PVSGather",
    "QueryTargets",
    "SetupPlatforms",
    "ValidatePlatforms",
    "WriteDocumentation",
    "WriteMetadata",
]


# fmt: off
@dataclass
class Arguments:
    """Unreal Build tools arguments

    Attributes
    ----------
    mode, str
        Select tool mode. One of the following (default tool mode is "Build"):
    verbose, bool
        Increase output verbosity
    veryverbose, bool
        Increase output verbosity more
    log, str
        Specify a log file location instead of the default Engine/Programs/UnrealBuildTool/Log.txt
    tracewrites, str
        Trace writes requested to the specified file
    timestamps, bool
        Include timestamps in the log
    frommsbuild, bool
        Format messages for msbuild
    progress, bool
        Write progress messages in a format that can be parsed by other programs
    nomutex, bool
        Allow more than one instance of the program to run at once
    waitmutex, bool
        Wait for another instance to finish and then start, rather than aborting immediately
    remoteini, str
        Remote tool ini directory
    clean, bool
        Clean build products. Equivalent to -Mode=Clean
    projectfiles, bool
        Generate project files based on IDE preference. Equivalent to -Mode=GenerateProjectFiles
    projectfileformat, str
        Generate project files in specified format. May be used multiple times.
    makefile, bool
        Generate Linux Makefile
    cmakefile, bool
        Generate project files for CMake
    qmakefile, bool
        Generate project files for QMake
    kdevelopfile, bool
        Generate project files for KDevelop
    codelitefiles, bool
        Generate project files for Codelite
    xcodeprojectfiles, bool
        Generate project files for XCode
    eddieprojectfiles, bool
        Generate project files for Eddie
    vscode, bool
        Generate project files for Visual Studio Code
    vsmac, bool
        Generate project files for Visual Studio Mac
    clion, bool
        Generate project files for CLion
    rider, bool
        Generate project files for Rider

    """
    mode                : str | None = choice(*modes, default="Build")  # Select tool mode. One of the following (default tool mode is "Build"):
    verbose             : bool = False  # Increase output verbosity
    veryverbose         : bool = False  # Increase output verbosity more
    log                 : str | None  = None  # Specify a log file location instead of the default Engine/Programs/UnrealBuildTool/Log.txt
    tracewrites         : bool = False  # Trace writes requested to the specified file
    timestamps          : bool = False  # Include timestamps in the log
    frommsbuild         : bool = False  # Format messages for msbuild
    progress            : bool = False  # Write progress messages in a format that can be parsed by other programs
    nomutex             : bool = False  # Allow more than one instance of the program to run at once
    waitmutex           : bool = False  # Wait for another instance to finish and then start, rather than aborting immediately
    remoteini           : bool = False  # Remote tool ini directory
    clean               : bool = False  # Clean build products. Equivalent to -Mode=Clean
    projectfiles        : bool = False  # Generate project files based on IDE preference. Equivalent to -Mode=GenerateProjectFiles
    projectfileformat   : str | None  = None  # Generate project files in specified format. May be used multiple times.
    makefile            : bool = False  # Generate Linux Makefile
    cmakefile           : bool = False  # Generate project files for CMake
    qmakefile           : bool = False  # Generate project files for QMake
    kdevelopfile        : bool = False  # Generate project files for KDevelop
    codelitefiles       : bool = False  # Generate project files for Codelite
    xcodeprojectfiles   : bool = False  # Generate project files for XCode
    eddieprojectfiles   : bool = False  # Generate project files for Eddie
    vscode              : bool = False  # Generate project files for Visual Studio Code
    vsmac               : bool = False  # Generate project files for Visual Studio Mac
    clion               : bool = False  # Generate project files for CLion
    rider               : bool = False  # Generate project files for Rider
# fmt: on


TARGETS = [
    "UnrealHeaderTool",
    "ShaderCompileWorker",
    "UnrealPak",
]


class UBT(Command):
    """Runs Unreal build tool as is.

    Notes
    -----
    Experimental do not use

    """

    name: str = "ubt"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, UBT)
        parser.add_argument("target", type=str, help="target name")
        parser.add_argument(
            "--project",
            default=None,
            type=str,
            help="Path to the project, example: <project>.uproject",
        )
        parser.add_argument(
            "--platform",
            default="Win64",
            type=str,
            help="Platform to compile for",
            choices=get_build_platforms(),
        )
        parser.add_argument(
            "--profile",
            default="Development",
            type=str,
            help="Build profile",
            choices=get_build_modes(),
        )
        parser.add_arguments(Arguments, dest="args")

    @staticmethod
    def execute(args):
        pargs = [
            ubt(),
            args.target,
            args.platform,
            args.profile,
        ]

        if args.project:
            project = find_project(args.project)
            pargs.append(f"-Project={project}")

        pargs += command_builder(asdict(args.args))

        return run(pargs, check=True).returncode


COMMANDS = UBT



from git import Repo

from uetools.core.command import Command, chdir, newparser
from uetools.core.conf import engine_root
from uetools.core.run import run


class Engine(Command):
    """Update the engine source code"""

    name: str = "engine"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, Engine)
        parser.add_argument(
            "--branch", type=str, default="5.0", help="checkout a branch"
        )
        parser.add_argument("--remote", type=str, default="origin", help="remote name")
        parser.add_argument(
            "--update",
            action="store_true",
            default=False,
            help="Pull the latest changes",
        )
        parser.add_argument(
            "--dry",
            action="store_true",
            default=False,
            help="Print the command that are going to be executed",
        )

    @staticmethod
    def execute(args):
        root = engine_root()

        if args.update:
            cmd = Engine.update(root, args.remote)
        else:
            cmd = Engine.checkout(root, args.branch, args.remote)

        with chdir(root):
            return Engine.execute_cmd(cmd, dry=args.dry)

    @staticmethod
    def update(root, remote):
        """Update current branch by pulling the latest changes"""
        engine_repo = Repo(root)
        cmd = ["git", "pull", remote, engine_repo.active_branch]
        return [cmd]

    @staticmethod
    def checkout(root, branch, remote):
        """Checkout a new branch"""
        engine_repo = Repo(root)

        if branch == engine_repo.active_branch:
            cmd = [["git", "pull", remote, branch]]

        else:
            cmd = [
                ["git", "fetch", remote, branch],
                ["git", "checkout", branch],
            ]

        return cmd

    @staticmethod
    def execute_cmd(cmds, dry=False):
        """Execute commands"""
        for cmd in cmds:
            print(" ".join(cmd))

            if not dry:
                return run(cmd, check=True).returncode

        return 0


COMMANDS = Engine



# https://docs.unrealengine.com/5.0/en-US/world-partition-in-unreal-engine/


from dataclasses import dataclass
from typing import Optional

from uetools.core.command import Command, command_builder, newparser
from uetools.core.conf import editor_cmd
from uetools.core.run import run


# fmt: off
@dataclass
class Arguments:
    """Convert a UE4 map using world partition"""
    project                 : str
    map                     : str
    SCCProvider             : Optional[str] = None  # Specifies which source control provider to use. To run without source control, specify -SCCProvider=None.
    Verbose                 : bool = False  # Displays verbose logging.
    ConversionSuffix        : bool = False  # Appends the _WP suffix to a converted map. This is useful when converting Levels for testing purposes while keeping the source Level intact.
    DeleteSourceLevel       : bool = False  # Deletes source Levels after conversion.
    ReportOnly              : bool = False  # Reports what would happen during the conversion. Does not do the conversion.
    GenerateIni             : bool = False  # Generates a default .ini conversion file for this map. Does not do the conversion.
    SkipStableGUIDValidation: bool = False  # Skips the unstable actor GUIDs validation process. Levels with unstable actor GUIDs will result in different conversion output when converting several times. Resaving the Level fixes this.
    OnlyMergeSubLevels      : bool = False  # Converts and merges Levels and Sublevels to One File Per Actor without World Partition. The converted Level can be used as a Level Instance in a World Partition Level.
    FoliageTypePath         : Optional[str] = None  # Extracts Foliage Types as Assets to the given path. Use if the Level contains embedded Foliage Types.
# fmt: on


class WorldPartition(Command):
    """Convert a UE4 map using world partition"""

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, WorldPartition)
        parser.add_arguments(Arguments, dest="args")

    @staticmethod
    def execute(args):
        cmd = [editor_cmd()] + command_builder(args)
        return run(cmd, check=True).returncode



import json
import os
from dataclasses import dataclass
from typing import Optional

from uetools.core.command import Command, newparser
from uetools.core.conf import find_project
from uetools.core.run import run


@dataclass
class Arguments:
    """Install a plugin to an unreal project.

    Attributes
    ----------
    name: str
        Name of the project to modify.

    url: str
        url of the plugin to install.

    destination:str
        installation directory (defaults to: ``$PROJECT_NAME/Plugins/``)

    submodule: bool
        install the plugin as a git submodule (defaults to: ``False``)

    Examples
    --------

    .. code-block::

        # This will install the plugin inside RTSGame/Plugins/
        #  it will download the repository on put it inside the RTSGame/Plugins/ folder
        uecli install RTSGame VoxelPlugin https://github.com/Phyronnaz/VoxelPlugin --enable

        # This will install the plugin inside RTSGame/Plugins/
        # it will execute the following command:
        #    - git submodule add https://github.com/Phyronnaz/VoxelPlugin Plugins/VoxelPlugin
        #
        uecli install RTSGame VoxelPlugin https://github.com/Phyronnaz/VoxelPlugin --enable --destination Plugins --submodule

        # disable the plugin
        uecli disable RTSGame VoxelPlugin
    """

    name: str
    url: str
    destination: Optional[str] = None
    submodule: bool = False


class Install(Command):
    """Install a plugin to an unreal project."""

    name: str = "install"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, Install)
        parser.add_argument("name", type=str, help="Project name")
        parser.add_argument("plugin", type=str, help="name of the plugin")
        parser.add_argument("url", type=str, help="repository url of the plugin")
        parser.add_argument(
            "--enable",
            action="store_true",
            help="Enable the plugin in the project settings",
        )
        parser.add_argument(
            "--destination",
            type=str,
            default="Plugins",
            help="Plugin destination, relative to the root of the project",
        )
        parser.add_argument(
            "--submodule",
            action="store_true",
            default=False,
            help="add the plugin as a git submodule",
        )
        parser.add_argument(
            "--force",
            action="store_true",
            default=False,
            help="submodule force",
        )

    @staticmethod
    def execute(args):
        project = find_project(args.name)
        folder = os.path.dirname(project)

        dest_relative = f"{args.destination}/{args.plugin}"
        dest = f"{folder}/{dest_relative}"

        errors = 0

        if not os.path.exists(dest):
            if args.submodule:
                force = ["--force"] if args.force else []
                cmd = (
                    ["git", "submodule", "add"]
                    + force
                    + ["--depth", "1", args.url, dest_relative]
                )
            else:
                cmd = ["git", "clone", "--depth", "1", args.url, dest_relative]

            print(" ".join(cmd))
            errors += run(cmd, check=True, cwd=folder).returncode

        else:
            print(f"Folder {dest} exists already, skipping installation")

        if args.enable:
            project = find_project(args.name)

            with open(project, encoding="utf-8") as project_file:
                project_conf = json.load(project_file)

            plugins = project_conf.get("Plugins", [])
            for plugin in plugins:

                if plugin.get("Name") == args.plugin:
                    plugin["Enabled"] = True
                    break

            else:
                # If Plugins array was missing, it was created above
                # insert it just in case
                project_conf["Plugins"] = plugins
                plugins.append(dict(Name=args.plugin, Enabled=True))

            with open(project, "w", encoding="utf-8") as project_file:
                json.dump(project_conf, project_file)

        return errors


COMMANDS = Install



"""Entry point for the command line interface"""
from __future__ import annotations

from argparse import SUPPRESS, Action

import simple_parsing

from uetools.commands import commands, find_command
from uetools.core.conf import BadConfig, select_engine_version


# pylint: disable=protected-access
# pylint: disable=too-few-public-methods
class HelpFormatter(simple_parsing.SimpleHelpFormatter):
    """Tweak the arrgument usage format to not show too many duplicates"""

    def __init__(self, prog, max_help_position, width=None):
        super().__init__(prog=prog, max_help_position=max_help_position, width=width)
        self.command = find_command(prog.split(" ")[-1])

    def _get_help_string(self, action: Action) -> str | None:
        help = super()._get_help_string(action=action)
        lines = help.split("\n", maxsplit=1)
        return lines[0]

    def add_usage(self, usage, actions, groups, prefix=None):
        if usage is None and self.command:
            examples = "\n       ".join(self.command.examples())

            usage = examples + "\n\ndescription:\n    " + self.command.help()

        super().add_usage(usage, actions, groups, prefix)

    class _Section:
        def __init__(self, formatter, parent, heading=None):
            self.formatter = formatter
            self.parent = parent
            self.heading = heading
            self.items = []

        def format_help(self):
            """Format the help section"""

            offset = 0
            # format the indented section
            if self.parent is not None:
                self.formatter._indent()

            join = self.formatter._join_parts

            frags = [func(*args) for func, args in self.items]

            item_help = join(frags[offset:])

            if self.parent is not None:
                self.formatter._dedent()

            # return nothing if the section was empty
            if not item_help:
                return ""

            # add the heading if the section was non-empty
            # pylint: disable=consider-using-f-string
            if self.heading is not SUPPRESS and self.heading is not None:
                current_indent = self.formatter._current_indent
                heading = "%*s%s:\n" % (current_indent, "", self.heading)
            else:
                heading = ""

            # join the section-initial newline, the heading and the help
            return join(["\n", heading, item_help, "\n"])


# pylint: disable=too-few-public-methods
class ArgumentParser(simple_parsing.ArgumentParser):
    """Force the argument parse to use the right formatter"""

    def _get_formatter(self):
        return HelpFormatter(prog=self.prog, max_help_position=45, width=None)
        # return BaseArgumentHelpFormatter()
        # return


def parse_args(argv):
    """Setup the argument parser for all supported commands"""
    parser = ArgumentParser()
    parser.add_argument(
        "--engine-version",
        nargs="?",
        type=str,
        default=None,
        help="Engine version to use if you have multiple installed",
    )

    subparsers = parser.add_subparsers(dest="command")

    for _, command in commands.items():
        command.arguments(subparsers)

    args = parser.parse_args(argv)
    if args.engine_version is not None:
        select_engine_version(args.engine_version)
        args.engine_version = None

    return args


def args(*a):
    """Utility to turn arguments into a list"""
    return a


def main(argv=None):
    """Entry point for the command line interface"""
    import sys

    try:
        parsed_args = parse_args(argv)
    except BadConfig:
        sys.exit(-1)

    cmd_name = parsed_args.command
    command = commands.get(cmd_name)

    if command is None:
        print(f"Action `{cmd_name}` not implemented")
        return

    returncode = command.execute(parsed_args)

    if returncode is None:
        return

    if returncode != 0:
        sys.exit(returncode)


if __name__ == "__main__":
    main()



from __future__ import annotations

import os
from dataclasses import dataclass

from uetools.core.command import Command, command_builder, newparser
from uetools.core.conf import find_project, get_build_modes, uat
from uetools.core.run import run

commands = [
    "AnalyzeThirdPartyLibs",
    "BenchmarkBuild",
    "BenchmarkOptions",
    "BuildCMakeLib",
    "BuildCommonTools",
    "BuildCookRun",
    "BuildDerivedDataCache",
    "BuildEditor",
    "BuildGame",
    "BuildHlslcc",
    "BuildPhysX",
    "BuildPlugin",
    "BuildServer",
    "BuildTarget",
    "BuildThirdPartyLibs",
    "CheckBalancedMacros",
    "CheckCsprojDotNetVersion",
    "CheckForHacks",
    "CheckPerforceCase",
    "CheckRestrictedFolders",
    "CheckTargetExists",
    "CheckXcodeVersion",
    "CleanAutomationReports",
    "CleanFormalBuilds",
    "CopySharedCookedBuild",
    "CopyUAT",
    "CreateComponentZips",
    "CreatePlatformExtension",
    "CryptoKeys",
    "DownloadJupiterBuild",
    "ExportMcpTemplates",
    "ExtractPaks",
    "FinalizeInstalledBuild",
    "FixPerforceCase",
    "FixupRedirects",
    "GenerateDSYM",
    "GeneratePlatformReport",
    "IPhonePackager",
    "ListMobileDevices",
    "ListThirdPartySoftware",
    "Localise",
    "Localize",
    "MegaXGE",
    "MemreportHelper",
    "OpenEditor",
    "ParseMsvcTimingInfo",
    "RebuildHLOD",
    "RebuildLightMaps",
    "RecordPerformance",
    "ReplaceAssetsUsingManifest",
    "ResavePackages",
    "StashTarget",
    "SyncBinariesFromUGS",
    "SyncDDC",
    "SyncDepotPath",
    "SyncProject",
    "UnrealBuildUtilDummyBuildCommand",
    "UnstashTarget",
    "UpdateLocalVersion",
    "UploadDDCToAWS",
    "WorldPartitionBuilder",
    "ZipProjectUp",
    "ZipUtils",
    "P4WriteConfig",
    "Build",
    "BuildGraph",
    "CleanTempStorage",
    "TempStorageTests",
    "MakeCookedEditor",
    "CleanDevices",
    "PublishUnrealAutomationTelemetry",
    "RunEditorTests",
    "RunUnreal",
    "RunUnrealTests",
    "TestGauntlet",
    "ExportIPAFromArchive",
    "MakeIPA",
    "WriteIniValueToPlist",
    "LowLevelTests.Automation",
    "RunLowLevelTests",
    "Turnkey",
]

# fmt: off
@dataclass
class UATArgs:
    """Common UAT arguments"""
    project                 : str           # Project path (required), i.e: -project=QAGame, -project=Samples/BlackJack/BlackJack.uproject, -project=D:/Projects/MyProject.uproject,
    verbose                 : bool = False  # Enables verbose logging
    veryverbose             : bool = False  # Enables very verbose logging
    submit                  : bool = False  # Allows UAT command to submit changes
    nosubmit                : bool = False  # Prevents any submit attempts
    nop4                    : bool = False  # Disables Perforce functionality {default if not run on a build machine}
    p4                      : bool = False  # Enables Perforce functionality {default if run on a build machine}
    list                    : bool = False  # Lists all available commands
    nokill                  : bool = False  # Does not kill any spawned processes on exit
    compile                 : bool = False  # Force all script modules to be compiled
    nocompile               : bool = False  # Do not attempt to compile any script modules - attempts to run with whatever is up to date
    ignorebuildrecords      : bool = False  # Ignore build records (Intermediate/ScriptModule/ProjectName.json) files when determining if script modules are up to date
    uselocalbuildstorage    : bool = False  # Allows you to use local storage for your root build storage dir {default of P:\\Builds {on PC} is changed to Engine\\Saved\\LocalBuilds}. Used for local testing.
    waitfordebugger         : bool = False  # Waits for a debugger to be attached, and breaks once debugger successfully attached.
    # help                  : bool = False  # Displays help

@dataclass
class BuildCookRunArgs:
    """Arguments for BuildCookRun command"""
    destsample              : bool = False  # Destination Sample name),
    foreigndest             : bool = False  # Foreign Destination
    targetplatform          : bool = False  # target platform for building, cooking and deployment (also -Platform)
    servertargetplatform    : bool = False  # target platform for building, cooking and deployment of the dedicated server (also -ServerPlatform)
    foreign                 : bool = False  # Generate a foreign uproject from blankproject and use that
    foreigncode             : bool = False  # Generate a foreign code uproject from platformergame and use that
    crashreporter           : bool = False  # true if we should build crash reporter
    cook                    : bool = False  # cookonthefly Determines if the build is going to use cooked data
    skipcook                : bool = False  # use a cooked build, but we assume the cooked data is up to date and where it belongs, implies -cook
    skipcookonthefly        : bool = False  # in a cookonthefly build, used solely to pass information to the package step
    clean                   : bool = False  # wipe intermediate folders before building
    unattended              : bool = False  # assumes no operator is present, always terminates without waiting for something.
    pak                     : bool = False  # generate a pak file
    iostore                 : bool = False  # generate I/O store container file(s)
    cook4iostore            : bool = False  # generate I/O store container file(s)
    zenstore                : bool = False  # save cooked output data to the Zen storage server
    makebinaryconfig        : bool = False  # generate optimized config data during staging to improve loadtimes
    signpak                 : bool = False  # sign the generated pak file with the specified key, i.e. -signpak=C:/Encryption.keys. Also implies -signedpak.
    prepak                  : bool = False  # attempt to avoid cooking and instead pull pak files from the network, implies pak and skipcook
    signed                  : bool = False  # the game should expect to use a signed pak file.
    PakAlignForMemoryMapping: bool = False  # The game will be set up for memory mapping bulk data.
    skippak                 : bool = False  # use a pak file, but assume it is already built, implies pak
    skipiostore             : bool = False  # override the -iostore commandline option to not run it
    stage                   : bool = False  # put this build in a stage directory
    skipstage               : bool = False  # uses a stage directory, but assumes everything is already there, implies -stage
    manifests               : bool = False  # generate streaming install manifests when cooking data
    createchunkinstall      : bool = False  # generate streaming install data from manifest when cooking data, requires -stage & -manifests
    skipencryption          : bool = False  # skips encrypting pak files even if crypto keys are provided
    archive                 : bool = False  # put this build in an archive directory
    build                   : bool = False  # True if build step should be executed
    noxge                   : bool = False  # True if XGE should NOT be used for building
    CookPartialgc           : bool = False  # while cooking clean up packages as we are done with them rather then cleaning everything up when we run out of space
    CookInEditor            : bool = False  # Did we cook in the editor instead of in UAT
    IgnoreCookErrors        : bool = False  # Ignores cook errors and continues with packaging etc
    nodebuginfo             : bool = False  # do not copy debug files to the stage
    separatedebuginfo       : bool = False  # output debug info to a separate directory
    MapFile                 : bool = False  # generates a *.map file
    nocleanstage            : bool = False  # skip cleaning the stage directory
    run                     : bool = False  # run the game after it is built (including server, if -server)
    cookonthefly            : bool = False  # run the client with cooked data provided by cook on the fly server
    Cookontheflystreaming   : bool = False  # run the client in streaming cook on the fly mode (don't cache files locally instead force reget from server each file load)
    fileserver              : bool = False  # run the client with cooked data provided by UnrealFileServer
    dedicatedserver         : bool = False  # build, cook and run both a client and a server (also -server)
    client                  : bool = False  # build, cook and run a client and a server, uses client target configuration
    noclient                : bool = False  # do not run the client, just run the server
    logwindow               : bool = False  # create a log window for the client
    package                 : bool = False  # package the project for the target platform
    skippackage             : bool = False  # Skips packaging the project for the target platform
    distribution            : bool = False  # package for distribution the project
    PackageEncryptionKeyFile: bool = False  # Path to file containing encryption key to use in packaging
    prereqs                 : bool = False  # stage prerequisites along with the project
    applocaldir             : bool = False  # location of prerequisites for applocal deployment
    Prebuilt                : bool = False  # this is a prebuilt cooked and packaged build
    AdditionalPackageOptions: bool = False  # extra options to pass to the platform's packager
    deploy                  : bool = False  # deploy the project for the target platform
    getfile                 : bool = False  # download file from target after successful run
    IgnoreLightMapErrors    : bool = False  # Whether Light Map errors should be treated as critical
    archivemetadata         : bool = False  # Archive extra metadata files in addition to the build (e.g. build.properties)
    createappbundle         : bool = False  # When archiving for Mac, set this to true to package it in a .app bundle instead of normal loose files
    iterativecooking        : bool = False  # Uses the iterative cooking, command line: -iterativecooking or -iterate
    CookMapsOnly            : bool = False  # Cook only maps this only affects usage of -cookall the flag
    CookAll                 : bool = False  # Cook all the things in the content directory for this project
    SkipCookingEditorContent: bool = False  # Skips content under /Engine/Editor when cooking
    FastCook                : bool = False  # Uses fast cook path if supported by target
    device                  : bool = False  # Devices to run the game on
    serverdevice            : bool = False  # Device to run the server on
    skipserver              : bool = False  # Skip starting the server
    nullrhi                 : bool = False  # add -nullrhi to the client commandlines
    fakeclient              : bool = False  # adds ?fake to the server URL
    editortest              : bool = False  # rather than running a client, run the editor instead
    RunAutomationTests      : bool = False  # when running -editortest or a client, run all automation tests, not compatible with -server
    ForceMonolithic         : bool = False  # Toggle to combined the result into one executable
    ForceDebugInfo          : bool = False  # Forces debug info even in development builds
    ForceNonUnity           : bool = False  # Toggle to disable the unity build system
    ForceUnity              : bool = False  # Toggle to force enable the unity build system
    Licensee                : bool = False  # If set, this build is being compiled by a licensee
    NoSign                  : bool = False  # Skips signing of code/content files.
    stagingdirectory        : str | None     = None  # Directory to copy the builds to, i.e. -stagingdirectory=C:/Stage
    unrealexe               : str | None     = None  # Name of the Unreal Editor executable, i.e. -unrealexe=UnrealEditor.exe
    archivedirectory        : str | None     = None  # Directory to archive the builds to, i.e. -archivedirectory=C:/Archive
    cmdline                 : str | None     = None  # command line to put into the stage in UECSCommandLine.txt
    bundlename              : str | None     = None  # string to use as the bundle name when deploying to mobile device
    map                     : str | None     = None  # map to run the game with
    AdditionalServerMapParams: str | None    = None  # Additional server map params, i.e ?param=value
    numclients              : int | None     = None  # Start extra clients, n should be 2 or more
    addcmdline              : str | None     = None  # Additional command line arguments for the program
    servercmdline           : str | None     = None  # Additional command line arguments for the program
    clientcmdline           : str | None     = None  # Override command line arguments to pass to the client
    Crash                   : int | None     = None  # when running -editortest or a client, adds commands like debug crash, debug rendercrash, etc based on index
    deviceuser              : str | None     = None  # Linux username for unattended key genereation
    devicepass              : str | None     = None  # Linux password
    RunTimeoutSeconds       : float | None   = None  # timeout to wait after we lunch the game
    SpecifiedArchitecture   : str | None     = None  # Determine a specific Minimum OS
    UbtArgs                 : str | None     = None  # extra options to pass to ubt
    MapsToRebuildLightMaps  : str | None     = None  # List of maps that need light maps rebuilding
    MapsToRebuildHLODMaps   : str | None     = None  # List of maps that need HLOD rebuilding


@dataclass
class BuildEditorArgs:
    """Builds the editor for the specified project. Example BuildEditor -project=QAGame.
    Note: Editor will only ever build for the current platform in a Development config and required tools will be included
    """
    notools: bool = False  # Don't build any tools (UHT, ShaderCompiler, CrashReporter


@dataclass
class LocalizeArgs:
    """Updates the external localization data using the arguments provided.

    Examples
    --------

    .. code-block:: console

       uecli uat Localize --project GamekitDev --IncludePlugins --ParallelGather --LocalizationSteps Gather --LocalizationProjectNames GamekitDev

       uecli local --project RTSGame --run GatherText --target RTSGame

    """
    UEProjectRoot                   : str | None  = None   # Optional root-path to the project we're gathering for (defaults to CmdEnv.LocalRoot if unset).",
    UEProjectDirectory              : str | None  = None   # Sub-path to the project we're gathering for (relative to UEProjectRoot).",
    UEProjectName                   : str | None  = None   # Optional name of the project we're gathering for (should match its .uproject file, eg QAGame).",
    LocalizationProjectNames        : str | None  = None   # Comma separated list of the projects to gather text from.",
    LocalizationBranch              : str | None  = None   # Optional suffix to use when uploading the new data to the localization provider.",
    LocalizationProvider            : str | None  = None   # Optional localization provide override."),
    LocalizationSteps               : str | None  = None   # Optional comma separated list of localization steps to perform [Download, Gather, Import, Export, Compile, GenerateReports, Upload] (default is all). Only valid for projects using a modular config.",
    IncludePlugins                  : bool        = False  # Optional flag to include plugins from within the given UEProjectDirectory as part of the gather. This may optionally specify a comma separated list of the specific plugins to gather (otherwise all plugins will be gathered).",
    ExcludePlugins                  : str | None  = None   # Optional comma separated list of plugins to exclude from the gather.",
    IncludePlatforms                : bool        = False  # Optional flag to include platforms from within the given UEProjectDirectory as part of the gather.",
    AdditionalCSCommandletArguments : str | None  = None   # Optional arguments to pass to the gather process.",
    ParallelGather                  : bool        = False  # Run the gather processes for a single batch in parallel rather than sequence.",
    OneSkyProjectGroupName          : str | None  = None
# fmt: on


class UAT(Command):
    """Runs Unreal Automation tool.

    Notes
    -----

    We recommend users to create their own command class that specialize its use to a single purpose.

    We try not to use UAT as it ends up calling the editor anyway.
    """

    name: str = "uat"

    @staticmethod
    def arguments(subparsers):
        """Defines UAT arguments"""
        parser = newparser(subparsers, UAT)

        parser.add_argument(
            "cmd", type=str, choices=commands, help="UAT Command to execute"
        )
        parser.add_argument(
            "--configuration",
            type=str,
            default="Development",
            choices=get_build_modes(),
            help="Build configuration",
        )
        parser.add_arguments(LocalizeArgs, dest="localize")
        parser.add_arguments(BuildEditorArgs, dest="build_editor")
        parser.add_arguments(BuildCookRunArgs, dest="build_cook_run")
        parser.add_arguments(UATArgs, dest="uat")

    @staticmethod
    def execute(args):
        """Execute the UAT command"""

        uat_cmd = vars(args).pop("cmd")

        uproject = find_project(args.uat.project)

        if uat_cmd.lower() in ("localize", "localise"):
            uproject_folder = os.path.dirname(uproject)
            project_name = os.path.basename(uproject_folder)

            if args.localize.UEProjectRoot is None:
                args.localize.UEProjectRoot = uproject_folder

            if args.localize.UEProjectDirectory is None:
                args.localize.UEProjectDirectory = ""

            if args.localize.UEProjectName is None:
                args.localize.UEProjectName = project_name

        args = command_builder(args)
        cmd = [uat()] + [uat_cmd] + args + ["-nocompileuat"]

        print(" ".join(cmd))
        return run(
            cmd,
            check=True,
        ).returncode


COMMANDS = UAT


@staticmethod
def execute_uat_test(args):
    """UAT & Gauntlet lookt totally unusable"""
    # Gauntlet seems to simply be launching the editor
    name = args.name
    test = args.test

    project = find_project(name)

    # """
    # Gauntlet.Automation:
    #     CleanDevices
    #     PublishUnrealAutomationTelemetry
    #     RunEditorTests
    #     RunUnreal
    #     RunUnrealTests
    #     TestGauntlet
    # """

    # Error: Test EditorTest.EditorTestNode threw an exception during launch. Skipping test. Ex: Unreal tests should use ApplyToConfig(Config, Role, OtherRoles)
    #    at Gauntlet.UnrealTestConfiguration.ApplyToConfig(UnrealAppConfig AppConfig) in E:\\UnrealEngine\\Engine\\Source\\Programs\\AutomationTool\\Gauntlet\\Unreal\\Base\\Gauntlet.UnrealTestConfiguration.cs:line 803
    #    at EditorTest.EditorTestConfig.ApplyToConfig(UnrealAppConfig AppConfig, UnrealSessionRole ConfigRole, IEnumerable`1 OtherRoles) in E:\\UnrealEngine\\Engine\\Source\\Programs\\AutomationTool\\Gauntlet\\Editor\\RunEditorTests.cs:line 53
    #    at Gauntlet.UnrealBuildSource.CreateConfiguration(UnrealSessionRole Role, IEnumerable`1 OtherRoles) in E:\\UnrealEngine\\Engine\\Source\\Programs\\AutomationTool\\Gauntlet\\Unreal\\BuildSource\\Gauntlet.UnrealBuildSource.cs:line 519
    #    at Gauntlet.UnrealSession.LaunchSession() in E:\\UnrealEngine\\Engine\\Source\\Programs\\AutomationTool\\Gauntlet\\Unreal\\Base\\Gauntlet.UnrealSession.cs:line 667
    #    at Gauntlet.UnrealTestNode`1.StartTest(Int32 Pass, Int32 InNumPasses) in E:\\UnrealEngine\\Engine\\Source\\Programs\\AutomationTool\\Gauntlet\\Unreal\\Base\\Gauntlet.UnrealTestNode.cs:line 702
    #    at Gauntlet.TextExecutor.StartTest(TestExecutionInfo TestInfo, Int32 Pass, Int32 NumPasses) in E:\\UnrealEngine\\Engine\\Source\\Programs\\AutomationTool\\Gauntlet\\Framework\\Gauntlet.TestExecutor.cs:line 553
    # Error: Test EditorTest.EditorTestNode (Win64 Development EditorGame) failed to start
    # Test EditorTest.EditorTestNode (Win64 Development EditorGame) NotStarted
    cmd_args = [
        uat(),
        "RunEditorTests",
        f"-project={project}",
        f"-testname={test}",
    ] + ["-nocompileuat"]

    # ERROR: Unable to find type uetools in assemblies. Namespaces= System.Linq.Enumerable+SelectArrayIterator`2[System.String,System.String].
    #  (see E:\\UnrealEngine\\Engine\\Programs\\AutomationTool\\Saved\\Logs\\Log.txt for full exception trace)
    cmd_args = [
        uat(),
        "RunUnreal",
        f"-project={project}",
        "-platform=Win64",
        "--configuration=Development",
        "-build=local",
        f"-test={args.test}",
        # f'-build={engine}',
        # f'-log={log_path}',
        # '-TestExit=Automation Test Queue Empty',
        # f'-ExecCmds=Automation RunTests {args.test}',
    ] + ["-nocompileuat"]

    # ERROR: Unable to find type uetools in assemblies. Namespaces= System.Linq.Enumerable+SelectArrayIterator`2[System.String,System.String].
    #  (see E:\\UnrealEngine\\Engine\\Programs\\AutomationTool\\Saved\\Logs\\Log.txt for full exception trace)
    cmd_args = [
        uat(),
        "RunUnrealTests",
        f"-project={project}",
        # f"-platform=Win64",
        # f"--configuration=Development",
        "-build=local",
        f"-test={args.test}",
        # f'-build={engine}',
        # f'-log={log_path}',
        # '-TestExit=Automation Test Queue Empty',
        # f'-ExecCmds=Automation RunTests {args.test}',
    ] + ["-nocompileuat"]

    run(cmd_args, check=True)



import os
from dataclasses import dataclass

import pkg_resources

from uetools.core.command import Command, newparser
from uetools.core.conf import find_project


@dataclass
class Arguments:
    """Create a dedicated server target for a given project

    Attributes
    ----------
    project: str
        Name of the project to add the server target

    Examples
    --------

    .. code-block:: console

       # Add the new server target
       uecli dedicated RTSGame

       # update visual studio project
       uecli regenerate RTSGame

       # Build the new target
       uecli build RTSGameServer

    """

    project: str


class Dedicated(Command):
    """Create a dedicated server target for a given project"""

    name: str = "dedicated"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, Dedicated)
        parser.add_argument(
            "project", default=None, type=str, help="name of your project"
        )

    @staticmethod
    def execute(args):
        project = find_project(args.project)
        folder = os.path.dirname(project)

        source_folder = os.path.join(folder, "Source")
        server_target = os.path.join(source_folder, f"{args.project}Server.Target.cs")

        if os.path.exists(server_target):
            print(f"{server_target} already exists")
            return 0

        Dedicated.generate_server_target(args.project, server_target)

        return 0

    @staticmethod
    def generate_server_target(project, server_target):
        """Add the UBT server target to a given project"""
        template = pkg_resources.resource_filename(
            __name__, "../templates/TemplateServer.Target"
        )

        with open(template, encoding="utf-8") as template_file:
            template = template_file.read()

        template = template.replace("{ProjectName}", project)

        with open(server_target, "w", encoding="utf-8") as file:
            file.write(template)


COMMANDS = Dedicated



import logging

from uetools.format.base import Formatter, colored, colors, short

log = logging.getLogger()


class CookingFormatter(Formatter):
    """Makes the cooking summary more readable"""

    def __init__(self, col=None) -> None:
        super().__init__(col)
        self.summary_starts = 0
        self.print_non_matching = True

    # pylint: disable=too-many-arguments,unused-argument
    def default_format(
        self, datetime=None, frame=None, category=None, verbosity=None, message=None
    ):
        """Formattings for the default log format"""
        verb = short.get(verbosity, " ")

        if verb == " ":
            log.debug("%s", verbosity)

        category = f"{{:<{self.longest_category}}}".format(category)

        color = colors.get(verb)

        print(f"[{verb}][{category}] {colored(message, color=color)}")

    # pylint: disable=too-many-arguments
    def format(
        self, datetime=None, frame=None, category=None, verbosity=None, message=None
    ):
        if "Warning/Error Summary (Unique only)" in message:
            self.summary_starts += 1
            self.default_format(datetime, frame, category, verbosity, message)
            return

        if "-----------------------------------" in message:
            self.summary_starts += 1
            self.default_format(datetime, frame, category, verbosity, message)
            return

        if message == 0:
            self.summary_starts = 0

        if self.summary_starts == 2:
            self.default_format(datetime, frame, category, verbosity, "> " + message)
            return

        self.default_format(datetime, frame, category, verbosity, message)



"""Defines constants and utility used throughout the project"""
import json
import logging
import os
import platform
import re
import subprocess

from appdirs import user_config_dir

NAME = "uecli"
AUTHOR = "uetools"
CONFIG = user_config_dir(NAME, AUTHOR)
CONFIGNAME = "loc.json"
LATEST_CONF = None
WINDOWS = platform.system().startswith("Windows")
OSX = platform.system().startswith("Darwin")
LINUX = not WINDOWS and not OSX


logger = logging.getLogger()


def get_build_modes():
    """Returns the build modes supported by UBT"""
    return ["Tests", "Debug", "Development", "Shipping"]


def build_platform_from_editor(editor_platform):
    """Maps editor target to build target"""
    if editor_platform.startswith("Windows"):
        return "Win64"

    return "Linux"


def guess_platform():
    """Try to guess the platform we are building for"""
    if WINDOWS:
        return "Win64"

    if OSX:
        return "Mac"

    return "Linux"


def guess_editor_platform():
    """Try to guess for which platform we are trying to build for"""
    if WINDOWS:
        return "Windows"

    # this is probably wrong
    if OSX:
        return "Mac"

    return "Linux"


def load_conf():
    """Loads the configuration from the config file"""
    # pylint: disable=global-statement
    global LATEST_CONF

    config = os.path.join(CONFIG, CONFIGNAME)
    os.makedirs(CONFIG, exist_ok=True)

    if not os.path.exists(config):
        return {}

    with open(config, encoding="utf-8") as conffile:
        conf = json.load(conffile)

    LATEST_CONF = conf
    return conf


def save_conf(conf):
    """Saves the given configuration to the config file"""
    config = os.path.join(CONFIG, CONFIGNAME)
    os.makedirs(CONFIG, exist_ok=True)

    with open(config, "w", encoding="utf-8") as conffile:
        json.dump(conf, conffile, indent=2)


def update_conf(**kwargs):
    """Updates the configuration with the given key-value pairs"""
    conf = load_conf()
    conf.update(kwargs)
    save_conf(conf)


def binary(name):
    """Returns the name of a binary for the current platform"""
    if WINDOWS:
        return name + ".exe"

    return name


def bash(name):
    """Returns the name of a script for the current platform"""
    if WINDOWS:
        return name + ".bat"

    # if OSX:
    #    return name + '.command'

    return name + ".sh"


def ubt():
    """Returns the path to the unreal build tools executable"""
    engine = engine_folder()

    if WINDOWS:
        return os.path.join(
            engine, "Binaries", "DotNET", "UnrealBuildTool", binary("UnrealBuildTool")
        )

    return os.path.join(engine, "Build", "BatchFiles", "Linux", "Build.sh")


def uat():
    """Returns the path to the unreal automation executable"""
    engine = engine_folder()

    if WINDOWS:
        return os.path.join(engine, "Build", "BatchFiles", bash("RunUAT"))

    return os.path.join(engine, "Build", "BatchFiles", bash("RunUAT"))


def uat_exec():
    """Returns the compiled automation tool"""
    engine = engine_folder()

    uat_directory = os.path.join(engine, "Binaries", "DotNet", "AutomationTool")
    uat_exec = "AutomationTool.dll"

    if WINDOWS:
        return os.path.join(uat_directory, uat_exec)

    raise NotImplementedError()


def dotnet():
    """Return the dotnet executable to run c# programs"""
    if WINDOWS:
        return "dotnet"

    raise NotImplementedError()


def editor():
    """Returns the path to the editor executable"""
    engine = engine_folder()

    if WINDOWS:
        return os.path.join(
            engine,
            "Binaries",
            "Win64",
            binary("UnrealEditor"),
        )

    return os.path.join(
        engine,
        "Binaries",
        "Linux",
        binary("UnrealEditor"),
    )


def editor_cmd():
    """Returns the path to the editor executable (cmd version)"""
    engine = engine_folder()

    if WINDOWS:
        return os.path.join(
            engine,
            "Binaries",
            "Win64",
            binary("UnrealEditor-Cmd"),
        )

    return os.path.join(
        engine,
        "Binaries",
        "Linux",
        binary("UnrealEditor-Cmd"),
    )


def editor_commandlet(project_path, command):
    """Returns editor arguments to run a commandlet"""
    return [editor_cmd()] + [
        project_path,
        "-NoLiveCoding",
        "-fullstdoutlogoutput",
        "-utf8output",
        "-nullrhi",
        "-nosplash",
        "-nosound",
        "-nopause",
        "-unattended",
        f"-run={command}",
    ]


# Supported platforms are listed here
#       UnrealEngine\Engine\Source\Programs\UnrealBuildTool\Platform
#
UBT_PLATFORMS = {
    "Android",
    "HoloLens",
    "IOS",
    "Linux",
    "LinuxArm64",
    "Mac",
    "TVOS",
    "Win64",
}


def get_build_platforms():
    """Returns the platforms supported by UBT"""
    return UBT_PLATFORMS


EDITOR_PLATOFRMS = {
    "Android",
    "Android_ASTC",
    "Android_DXT",
    "Android_ETC2",
    "AndroidClient",
    "Android_ASTCClient",
    "Android_DXTClient",
    "Android_ETC2Client",
    "Android_Multi",
    "Android_MultiClient",
    "HoloLens",
    "HoloLensClient",
    "Windows",
    "WindowsEditor",
    "WindowsServer",
    "WindowsClient",
    "LinuxArm64",
    "LinuxArm64Server",
    "LinuxArm64Client",
    "Linux",
    "LinuxServer",
    "LinuxClient",
    "LinuxEditor",
}


def get_editor_platforms():
    """returns the platforms supported by the editor"""
    return EDITOR_PLATOFRMS


def find_project(name):
    """Returns the path to the project file

    Examples
    --------

    .. code-block:: python

       find_project('RTSGame')
       /op/project/RTSGame/RTSGame.uproject

       # if the game is inside the engine repository
       find_project('RTSGame')
       /op/UnrealEngine/RTSGame/RTSGame.uproject

       find_project('/op/project/RTSGame/RTSGame.uproject')
       /op/project/RTSGame/RTSGame.uproject

       # project name is extracted from target
       find_project('RTSGameEditor')
       /op/project/RTSGame/RTSGame.uproject

       find_project('RTSGameServer')
       /op/project/RTSGame/RTSGame.uproject

    """

    if os.path.isabs(name) and name.endswith(".uproject"):
        return name

    if name.endswith(".uproject"):
        name = name[:-9]

    if name.endswith("Editor"):
        name = name[:-6]

    elif name.endswith("Server"):
        name = name[:-6]

    elif name.endswith("Client"):
        name = name[:-6]

    folder = os.path.join(project_folder(), name)
    uproject_project = os.path.join(folder, f"{name}.uproject")

    if os.path.exists(uproject_project):
        return uproject_project

    uproject_engine = os.path.join(engine_root(), name, f"{name}.uproject")

    if os.path.exists(uproject_engine):
        return uproject_engine

    raise RuntimeError(f"None of {uproject_project}, {uproject_engine} exist")


class ConfigurationError(Exception):
    """Raised when uetools has not been configured yet"""


def project_folder():
    """Returns the folder user to store projects"""
    p = load_conf().get("project_path")

    if p is None:
        logger.warning(
            "Project folder is not defined, using UnrealEngine root as default"
        )
        return engine_root()

    return p


PRECOMPILED_UE_VERSION = re.compile(r"UE_(?P<Major>[0-9]*)\.(?P<Minor>\d{1,})")


def find_engine(installfolder):
    """Look for the precompiled UE version people install from the Launcher"""
    if not os.path.exists(installfolder):
        print("Could not find UnrealEngine installation path")
        return None

    folders = os.listdir(installfolder)
    found = []
    for folder in folders:
        if PRECOMPILED_UE_VERSION.match(folder):
            found.append(folder)

    if found:
        print("Found UnrealEngine installation...")

        for folder in found:
            print(" - ", installfolder + folder)

        return [os.path.join(installfolder, folder) for folder in found]

    print("Found installation path {installfolder} but no Engine was found")
    return []


def guess_engine_folder():
    """Try to guess the engine location by looking at standard locations"""
    default_paths = ["C:/Program Files/Epic Games/"]

    engines = []
    for path in default_paths:
        engines.extend(find_engine(path))

    raise RuntimeError("Not implemented")


SELECTED_VERSION = None


class BadConfig(Exception):
    pass


def get_selected_engine_version():
    return SELECTED_VERSION


VERSION_FILE_PATH = "/Source/Runtime/Launch/Resources/Version.h"


def retrieve_exact_engine_version(engine_folder, default=None, ignore_patch=True):
    """Retrieve the engine version from the headers"""
    path = engine_folder + VERSION_FILE_PATH

    if not os.path.exists(path):
        return default

    version_names = [
        "ENGINE_MAJOR_VERSION",
        "ENGINE_MINOR_VERSION",
        "ENGINE_PATCH_VERSION",
    ]

    versions = ["0", "0", "0"]

    n = len(version_names[0])
    i = 0

    with open(path) as f:
        for line in f.readlines():
            if version_names[i] in line:
                versions[i] = line.split(version_names[i])[-1].strip()
                i += 1

            if i >= len(version_names):
                break

    if ignore_patch:
        versions[-1] = "0"

    return ".".join(versions)


def select_engine_version(version):
    global SELECTED_VERSION

    SELECTED_VERSION = version

    p = load_conf().get("engines").get(version, None)

    if p is None:
        print(
            f"Engine version {version} does not exist, call first:\n\t"
            f"`uecli engine-add --version {version} --engine <...>`"
        )
        raise BadConfig()


def engine_folder(version=None):
    """returns the engine folder ``UnrealEngine/Engine``"""

    if SELECTED_VERSION is None:
        p = load_conf().get("engine_path")
    else:
        p = load_conf().get("engines").get(SELECTED_VERSION, None)

    if p is None:
        raise ConfigurationError(
            "Engine path is missing; call `uecli init --engine <...> --projects <...>` first"
        )

    return p


def engine_root():
    """returns the engine root ``UnrealEngine``"""
    return os.path.abspath(os.path.join(engine_folder(), ".."))


def ready():
    """Returns true if uetools was initialized"""
    return load_conf().get("engine_path") is not None


def get_version_tag(path, default="1.0"):
    """Retrieve a version tag using git"""
    try:
        cmd = "git --no-optional-locks describe --tags --abbrev=0"
        return subprocess.check_output(cmd.split(" "), cwd=path).decode("utf-8").strip()
    except Exception as err:
        print(f"Could not use git to find version for {path}")
        return default



import os
from dataclasses import asdict, dataclass

from simple_parsing import choice

from uetools.core.command import Command, command_builder, newparser
from uetools.core.conf import find_project, ubt
from uetools.core.run import run

NONE = object()

generators = [
    "Make",
    "CMake",
    "QMake",
    "KDevelop",
    "CodeLite",
    "VisualStudio",
    "VisualStudio2019",
    "VisualStudio2022",
    "XCode",
    "Eddie",
    "VisualStudioCode",
    "VisualStudioMac",
    "CLion",
    "Rider",
    NONE,
]

# fmt: off
@dataclass
class Arguments:
    """Generate project files

    Notes
    -----

    While UBT shows all those generator options, none of them really work.
    For them to be used you need to modify the ``BuildConfiguration.xml`` which is the configuration
    used for the project file configuration.

    The build configuration can be found in ``${ENGINE_ROOT}/Engine/Saved/UnrealBuildTool/BuildConfiguration.xml

    .. code-block:: xml

       <?xml version="1.0" encoding="utf-8" ?>
       <Configuration xmlns="https://www.unrealengine.com/BuildConfiguration">
           <ProjectFileGenerator>
               <Format>VisualStudioCode</Format>
           </ProjectFileGenerator>
       </Configuration>

    """

    makefile            : bool = False  # Generate Linux Makefile
    cmakefile           : bool = False  # Generate project files for CMake
    qmakefile           : bool = False  # Generate project files for QMake
    kdevelopfile        : bool = False  # Generate project files for KDevelop
    codelitefiles       : bool = False  # Generate project files for Codelite
    xcodeprojectfiles   : bool = False  # Generate project files for XCode
    eddieprojectfiles   : bool = False  # Generate project files for Eddie
    vscode              : bool = False  # Generate project files for Visual Studio Code
    vsmac               : bool = False  # Generate project files for Visual Studio Mac
    clion               : bool = False  # Generate project files for CLion
    rider               : bool = False  # Generate project files for Rider
    projectfiles        : bool = True   # Generate project files based on IDE preference.
    projectfileformat   : str = choice(*generators, default=NONE, type=str)
    game                : bool = True
    engine              : bool = True
    progress            : bool = True
# fmt: on


class Generate(Command):
    """Generate project files"""

    name: str = "regenerate"

    @staticmethod
    def arguments(subparsers):
        """Adds the arguments for this command to the given parser"""
        parser = newparser(subparsers, Generate)
        parser.add_argument("project", type=str, help="project name")
        parser.add_arguments(Arguments, dest="args")

    @staticmethod
    def execute(args):

        project = find_project(args.project)

        cmd = [ubt(), f"-project={project}", "-Mode=GenerateProjectFiles"]

        cmd_options = command_builder(asdict(args.args))

        cmd = cmd + cmd_options

        print(" ".join(cmd))

        root = os.path.dirname(project)

        return run(
            cmd,
            check=True,
            cwd=root,
        ).returncode


COMMANDS = Generate



from uetools.format.base import Formatter, colored


class TestFormatter(Formatter):
    """Format test output to be more readable"""

    def __init__(self, col=None) -> None:
        super().__init__(col)
        self.indent = 0
        self.allow_everything = False
        self.iterating_overlist = False

    def default_format(self, *args, **kwargs):
        """Default format function"""
        Formatter.format(self, *args, **kwargs)

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-return-statements
    def format(
        self, datetime=None, frame=None, category=None, verbosity=None, message=None
    ):
        """Format test output to be more readable"""
        valid_line = (
            "Automation" in category or "Python" in category or self.allow_everything
        )

        if not valid_line:
            self.default_format(datetime, frame, category, verbosity, message)
            return

        # ------------
        # Testing init
        # ------------
        if "automation tests based on" in message:
            self.iterating_overlist = True
            self.default_format(datetime, frame, category, verbosity, "- " + message)
            return

        # This is only printed in logs
        if "Sending StopTests" in message:
            self.iterating_overlist = False
            self.default_format(datetime, frame, category, verbosity, message)
            return

        # -----
        # Tests
        # -----
        if "Test Started" in message:
            message = ">" + colored(message, "blue", attrs=["bold"])
            self.indent += 4
            self.iterating_overlist = False
            self.allow_everything = True
            self.default_format(datetime, frame, category, verbosity, message)
            return

        if "Test Completed" in message:
            self.indent -= 4
            self.allow_everything = False

            if "Success" in message:
                message = "<" + colored(message, "green")
            else:
                message = "<" + colored(message, "red")

            self.default_format(datetime, frame, category, verbosity, message)
            return

        if "BeginEvents" in message:
            self.indent += 4
            self.allow_everything = True
            self.default_format(datetime, frame, category, verbosity, "+ " + message)
            return

        if "EndEvents" in message:
            self.indent -= 4
            self.allow_everything = False
            self.default_format(datetime, frame, category, verbosity, "- " + message)
            self.default_format(datetime, frame, category, verbosity, "")
            return

        if self.iterating_overlist:
            self.default_format(
                datetime, frame, category, verbosity, "    * " + message.strip()
            )
            return

        message = " " * self.indent + message.strip()
        self.default_format(datetime, frame, category, verbosity, message)



from dataclasses import dataclass
from typing import Optional

from uetools.core.command import Command, newparser
from uetools.core.conf import editor, find_project
from uetools.core.run import popen_with_format
from uetools.format.base import Formatter


# This is not used technically, but we keep it for consistency with the other commands
# also help with the doc generation as this object appears on top
@dataclass
class Arguments:
    """Open the editor for a given project

    Attributes
    ----------
    project: str
        Name of the project to serve

    map: str
        Name of the map to serve

    dedicated: bool
        If true starts a dedicated server, otherwise a listen server (one local player that can host remote players)

    Examples
    --------

    .. code-block:: console

       uecli server RTSGame

    """

    # project: str
    # map: str
    dedicated: bool = False  # If true will start a dedicated server, otherwise a listen server (one local player that can host remote players)
    port: int = 8123  # Server port


@dataclass
class MapParameters:
    """Parameters added to the Map URL"""

    bIsLanMatch: bool = False
    bIsFromInvite: bool = False
    spectatoronly: bool = False
    gameinfo: Optional[str] = None


class Server(Command):
    """Launch the editor as a server"""

    name: str = "server"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, Server)

        # this makes it ugly
        # editor.add_arguments(Arguments, dest="server")

        parser.add_argument(
            "project", metavar="project", type=str, help="Name of the project to serve"
        )
        parser.add_argument(
            "map",
            metavar="map",
            type=str,
            help=" Name of the map to serve (if the map is located inside the map folder, just the name of the map is needed,"
            "if not the full path is needed including the extension, e.g. /Game/NotMap/MyMap.umap)",
        )
        parser.add_arguments(Arguments, dest="args")
        parser.add_arguments(MapParameters, dest="params")
        parser.add_argument(
            "--dry",
            action="store_true",
            default=False,
            help="Print the command it will execute without running it",
        )

    @staticmethod
    def execute(args):
        project = find_project(args.project)

        cmd = [
            editor(),
            project,
        ]

        map_options = []

        if not args.args.dedicated:
            map_options.append("?listen")

        if args.params.bIsLanMatch:
            map_options.append("?bIsLanMatch=1")

        if args.params.bIsFromInvite:
            map_options.append("?bIsFromInvite=1")

        if args.params.spectatoronly:
            map_options.append("?spectatoronly")

        if args.params.gameinfo:
            map_options.append(f"?game={args.params.gameinfo}")

        mapname = args.map + "&".join(map_options)
        cmd.append(mapname)

        cmd.append(f"-port={args.args.port}")
        cmd.append("-game")

        if args.args.dedicated:
            cmd.append("-server")

        cmd.append("-FullStdOutLogOutput")
        print(" ".join(cmd))

        if not args.dry:
            fmt = Formatter()
            return popen_with_format(fmt, cmd)

        return 0


COMMANDS = Server



import json
import os

from uetools.core.command import Command, newparser
from uetools.core.conf import find_project
from uetools.core.ini import UnrealINIParser


class VSCode(Command):
    """Tweak your VSCode setting for this project to find python stub generated by Unreal Engine.
    This will enable you to get autocomplete for Unreal Engine classes and functions.

    Attributes
    ----------
    name: str
        project name

    yes: bool
        No user prompts (assume yes)

    Notes
    -----

    It tweaks both your project settings and your vs code settings.

    """

    name: str = "vscode"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, VSCode)
        parser.add_argument("name", type=str, help="Project name")
        parser.add_argument(
            "--yes", action="store_true", help="No user prompts (assume yes)"
        )

    @staticmethod
    def execute(args):
        VSCode.enable_python_editor(args)
        VSCode.add_stub_to_path(args)

        return 0

    @staticmethod
    def enable_python_editor(args):
        """Modify your project settings to enable python scripting in your project"""
        project = find_project(args.name)
        folder = os.path.dirname(project)

        conf = os.path.join(folder, "Config")
        default_engine = os.path.join(conf, "DefaultEngine.ini")

        with open(default_engine, encoding="utf-8") as file:
            config = UnrealINIParser(file)

        python_section = "/Script/PythonScriptPlugin.PythonScriptPluginUserSettings"
        config.insert(python_section, "bDeveloperMode", "True")
        config.insert(python_section, "bEnableContentBrowserIntegration", "True")

        with open(default_engine, "w", encoding="utf-8") as file:
            config.write(file)

    @staticmethod
    def add_stub_to_path(args):
        """Modify vscode settings to add the python stub path to the autocomplete path."""
        name = args.name

        project = find_project(name)
        folder = os.path.dirname(project)

        vscode_folder = os.path.join(folder, ".vscode")
        vscode_settings = os.path.join(vscode_folder, "settings.json")

        if not os.path.exists(vscode_settings):
            create_file = "Y" if args.yes else None

            while create_file not in ("Y", "N"):
                create_file = input(
                    ".vscode/settings.json does not exist, create ? (Y/N):"
                )
                create_file = create_file.upper()

            if create_file == "Y":
                os.makedirs(vscode_folder, exist_ok=True)
                with open(vscode_settings, "w", encoding="utf-8") as file:
                    file.write("{}")
            else:
                print("Nothing to do")
                return

        with open(vscode_settings, encoding="utf-8") as file:
            vssetting = json.load(file)

        autocomplete_key = "python.autocomplete.extraPaths"
        analysis_key = "python.analysis.extraPaths"

        extra_paths = vssetting.get(autocomplete_key, [])
        extra_paths.append(os.path.join(folder, "Intermediate", "PythonStub"))
        extra_paths = list(set(extra_paths))
        vssetting[autocomplete_key] = extra_paths

        extra_paths = vssetting.get(analysis_key, [])
        extra_paths.append(os.path.join(folder, "Intermediate", "PythonStub"))
        extra_paths = list(set(extra_paths))
        vssetting[analysis_key] = extra_paths

        with open(vscode_settings, "w", encoding="utf-8") as file:
            json.dump(vssetting, file, indent=2)


COMMANDS = VSCode



import os
import xml.etree.ElementTree as ET
from copy import deepcopy
from dataclasses import asdict, dataclass, field

from uetools.core.command import Command, newparser
from uetools.core.conf import engine_folder

valid_file = """
<?xml version="1.0" encoding="utf-8" ?>
<Configuration xmlns="https://www.unrealengine.com/BuildConfiguration">
    <ParallelExecutor>
        <MaxProcessorCount>32</MaxProcessorCount>
    </ParallelExecutor>
    <BuildConfiguration>
    </BuildConfiguration>
    <WindowsPlatform>
    </WindowsPlatform>
</Configuration>
"""


# Can I make a Dataclass of the UBT config and initialize it using the XML
# and save the XML from the dataclass ?
# 100% it is possible, just might take some time
# I can just save the dataclass as a dict and save the values as
#
# <dict>
#   <key>value</key>
# </dict>


@dataclass
class WindowsPlatformT:
    """Windows Platform"""


@dataclass
class ModuleConfigurationT:
    """Module Configuration"""


@dataclass
class FASTBuildT:
    """FAST Build"""


@dataclass
class TaskExecutorT:
    """TaskExecutor"""


@dataclass
class HybridExecutorT:
    """HybridExecutor"""


@dataclass
class ParallelExecutorT:
    """ParallelExecutor"""

    bStopCompilationAfterErrors: bool = False
    bShowCompilationTimes: bool = False
    bShowPerActionCompilationTimes: bool = True
    bLogActionCommandLines: bool = True
    bPrintActionTargetNames: bool = True
    MaxProcessorCount: int = 32
    ProcessorCountMultiplier: float = 1
    MemoryPerActionBytes: int = 0


@dataclass
class SNDBST:
    """SNDBS"""


@dataclass
class XGET:
    """XGE"""


@dataclass
class BuildModeT:
    """BuildMode"""


@dataclass
class ProjectFileGeneratorT:
    """ProjectFileGenerator"""


@dataclass
class HoloLensPlatformT:
    """HoloLensPlatform"""


@dataclass
class IOSToolChainT:
    """IOSToolChain"""


@dataclass
class WindowsTargetRulesT:
    """WindowsTargetRules"""


@dataclass
class CLionGeneratorT:
    """CLionGenerator"""


@dataclass
class CMakeFileGeneratorT:
    """CMakeFileGenerator"""


@dataclass
class CodeLiteGeneratorT:
    """CodeLiteGenerator"""


@dataclass
class EddieProjectFileGeneratorT:
    """EddieProjectFileGenerator"""


@dataclass
class KDevelopGeneratorT:
    """KDevelopGenerator"""


@dataclass
class MakefileGeneratorT:
    """MakefileGenerator"""


@dataclass
class QMakefileGeneratorT:
    """QMakefileGenerator"""


@dataclass
class RiderProjectGeneratorT:
    """RiderProjectGenerator"""


@dataclass
class VSCodeProjectFileGeneratorT:
    """VSCodeProjectFileGenerator"""


@dataclass
class VCMakeProjectFileGeneratorT:
    """VCMakeProjectFileGenerator"""


@dataclass
class VCProjetFileGeneratorT:
    """VCProjetFileGenerator"""


@dataclass
class XCodeProjectFileGeneratorT:
    """XCodeProjectFileGenerator"""


@dataclass
class SourceFileWorkingSetT:
    """SourceFileWorkingSet"""


@dataclass
class RemoteMacT:
    """RemoteMac"""


@dataclass
class LogT:
    """Log"""


@dataclass
class Configuration:
    """Configuration"""

    WindowsPlatform: WindowsPlatformT = field(default_factory=WindowsPlatformT)
    ModuleConfiguration: ModuleConfigurationT = field(
        default_factory=ModuleConfigurationT
    )
    FASTBuild: FASTBuildT = field(default_factory=FASTBuildT)
    TaskExecutor: TaskExecutorT = field(default_factory=TaskExecutorT)
    HybridExecutor: HybridExecutorT = field(default_factory=HybridExecutorT)
    ParallelExecutor: ParallelExecutorT = field(default_factory=ParallelExecutorT)
    SNDBS: SNDBST = field(default_factory=SNDBST)
    XGE: XGET = field(default_factory=XGET)
    BuildMode: BuildModeT = field(default_factory=BuildModeT)
    ProjectFileGenerator: ProjectFileGeneratorT = field(
        default_factory=ProjectFileGeneratorT
    )
    HoloLensPlatform: HoloLensPlatformT = field(default_factory=HoloLensPlatformT)
    IOSToolChain: IOSToolChainT = field(default_factory=IOSToolChainT)
    WindowsTargetRules: WindowsTargetRulesT = field(default_factory=WindowsTargetRulesT)
    CLionGenerator: CLionGeneratorT = field(default_factory=CLionGeneratorT)
    CMakeFileGenerator: CMakeFileGeneratorT = field(default_factory=CMakeFileGeneratorT)
    CodeLiteGenerator: CodeLiteGeneratorT = field(default_factory=CodeLiteGeneratorT)
    EddieProjectFileGenerator: EddieProjectFileGeneratorT = field(
        default_factory=EddieProjectFileGeneratorT
    )
    KDevelopGenerator: KDevelopGeneratorT = field(default_factory=KDevelopGeneratorT)
    MakefileGenerator: MakefileGeneratorT = field(default_factory=MakefileGeneratorT)
    QMakefileGenerator: QMakefileGeneratorT = field(default_factory=QMakefileGeneratorT)
    RiderProjectGenerator: RiderProjectGeneratorT = field(
        default_factory=RiderProjectGeneratorT
    )
    VSCodeProjectFileGenerator: VSCodeProjectFileGeneratorT = field(
        default_factory=VSCodeProjectFileGeneratorT
    )
    VCMakeProjectFileGenerator: VCMakeProjectFileGeneratorT = field(
        default_factory=VCMakeProjectFileGeneratorT
    )
    VCProjetFileGenerator: VCProjetFileGeneratorT = field(
        default_factory=VCProjetFileGeneratorT
    )
    XCodeProjectFileGenerator: XCodeProjectFileGeneratorT = field(
        default_factory=XCodeProjectFileGeneratorT
    )
    SourceFileWorkingSet: SourceFileWorkingSetT = field(
        default_factory=SourceFileWorkingSetT
    )
    RemoteMac: RemoteMacT = field(default_factory=RemoteMacT)
    Log: LogT = field(default_factory=LogT)


def find_or_insert(root, *paths, namespaces=None):
    node = root
    for i, p in enumerate(paths):
        full_path = "ue:" + "/ue:".join(paths[: i + 1])
        val = root.find(full_path, namespaces)

        if val is None:
            val = ET.Element(p)
            node.insert(0, val)

        node = val

    return val


class LimitCPU(Command):
    """Disable unused plugin that are loading by default"""

    name: str = "limitcpu"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, LimitCPU)
        parser.add_argument(
            "--cpu", default=32, type=int, help="Limit the number of cpu to use"
        )

    @staticmethod
    def execute(args):
        #
        # https://docs.unrealengine.com/4.27/en-US/ProductionPipelines/DevelopmentSetup/BuildConfigurations/
        # https://docs.unrealengine.com/5.0/en-US/build-configuration-for-unreal-engine/
        #
        engine = engine_folder()
        namespaces = {"ue": "https://www.unrealengine.com/BuildConfiguration"}
        global_ubt_config = os.path.join(
            engine, "Saved", "UnrealBuildTool", "BuildConfiguration.xml"
        )
        # user_ubt_config = os.path.join(home, "AppData", "Roaming", "Unreal Engine", "UnrealBuildTool", "BuildConfiguration.xml")
        # local_ubt_config = os.path.join(home, "Documents", "Unreal Engine", "UnrealBuildTool", "BuildConfiguration.xml")

        ET.register_namespace("", "https://www.unrealengine.com/BuildConfiguration")
        tree = ET.parse(global_ubt_config)

        root = tree.getroot()
        node = find_or_insert(
            root, "ParallelExecutor", "MaxProcessorCount", namespaces=namespaces
        )

        oldval = node.text
        node.text = str(args.cpu)

        print(f"CPU {oldval} => {args.cpu}")

        ET.indent(tree, space="  ", level=0)
        tree.write(global_ubt_config, xml_declaration=True, encoding="utf-8")
        return 0


COMMANDS = LimitCPU


def from_xml(filename: str) -> Configuration:
    """Parse UBT XML Configuration file"""
    namespaces = {"ue": "https://www.unrealengine.com/BuildConfiguration"}
    ET.register_namespace("", "https://www.unrealengine.com/BuildConfiguration")
    tree = ET.parse(filename)

    root = tree.getroot()
    config = asdict(Configuration())

    path = []
    _from_xml(config, path, root, namespaces, 0)

    return config


def _from_xml(config, path: list, tree, namespaces, depth):
    configdict = asdict(config)

    for k, v in configdict.items():
        fullpath = deepcopy(path) + [k]
        ref = getattr(config, k)

        if isinstance(v, dict):
            _from_xml(ref, fullpath, tree, namespaces, depth + 1)

        else:
            fullpath = "/".join([f"ue:{p}" for p in fullpath])
            node = tree.find(fullpath, namespaces)
            setattr(config, k, node.text)


def to_xml(config: Configuration) -> str:
    """Writes a UBT XML configuration file"""
    frags = []

    dictconf = asdict(config)

    frags.append('<?xml version="1.0" encoding="utf-8" ?>\n')
    frags.append(
        '<Configuration xmlns="https://www.unrealengine.com/BuildConfiguration">'
    )
    _to_xml(dictconf, frags, 1)
    frags.append("</Configuration>")

    return "".join(frags)


def _to_xml(dictionary: dict, output: list, depth: int) -> None:
    idt = "  " * depth

    for k, v in dictionary.items():
        if not v:
            continue

        if isinstance(v, dict):
            output.append(f"\n{idt}<{k}>")
            _to_xml(v, output, depth + 1)
            output.append(f"\n{idt}</{k}>\n")
        else:
            output.append(f"\n{idt}<{k}>")
            output.append(f"{v}")
            output.append(f"</{k}>")


def check():
    """check POC"""
    config = Configuration()
    print(to_xml(config))


if __name__ == "__main__":
    check()



"""Argument formatting experimentation"""
from __future__ import annotations

from argparse import Action, _SubParsersAction
from dataclasses import dataclass, field
from functools import singledispatch


class BaseArgumentHelpFormatter:
    def _format_args(self, action, default_metavar):
        # print("_format_args", args, kwargs)
        return ""

    def add_usage(self, usage, actions, groups, prefix=None):
        print("add_usage", usage, actions, groups, prefix)
        return ""

    def format_help(self):
        print("format_help")
        return ""

    # Argument groups
    # start_section
    #  add_text
    #  add_arguments
    # end_section

    def start_section(self, heading: str) -> None:
        print("start_section", heading)
        return ""

    def add_text(self, text: str) -> None:
        print("add_text", text)
        return ""

    def add_arguments(self, actions: list[Action]) -> None:
        print("add_arguments", actions)
        return ""

    def end_section(self) -> None:
        print("end_section")
        return ""


@singledispatch
def format_action(action, level):
    return repr(action)


def _(action: _SubParsersAction):

    frags = []

    for name, _ in action.choices.items():
        helpstr = ""
        frags.append((name, helpstr))

    return


@dataclass
class ArgumentSection:
    name: str = None
    description: str = None
    actions: list[Action] = field(default_factory=list)
    level: int = 0

    def __repr__(self):
        idt = "  "
        idt1 = idt * self.level
        frags = [
            f"{idt1}{self.name}: {self.description}",
        ] + [f"{idt1}{idt}{a}" for a in self.actions]

        return "\n".join(frags)


class ArgumentHelpFormatter(BaseArgumentHelpFormatter):
    def __init__(self):
        self.sections = []
        self.generated = []

    def _format_args(self, action, default_metavar):
        # print("_format_args", args, kwargs)
        return

    def add_usage(self, usage, actions, groups, prefix=None):
        print("add_usage", usage, actions, groups, prefix)
        return ""

    def format_help(self):
        print("format_help")
        frags = []
        for g in self.generated:
            frags.append(repr(g))

        return "\n\n".join(frags)

    def start_section(self, heading: str) -> None:
        self.sections.append(ArgumentSection(name=heading, level=len(self.sections)))

    def add_text(self, text: str) -> None:
        if self.sections:
            self.sections[-1].description = text

    def add_arguments(self, actions: list[Action]) -> None:
        if self.sections:
            self.sections[-1].actions = actions

    def end_section(self) -> None:
        if self.sections:
            self.generated.append(self.sections.pop())



from __future__ import annotations

from dataclasses import dataclass

from uetools.core.command import Command, command_builder, newparser
from uetools.core.conf import editor_cmd, find_project
from uetools.core.run import popen_with_format
from uetools.format.base import Formatter

EDITOR_COMMANDSS = [
    # "debug {0}"
    "automation list",
    "runtests",
    "runall",
]

# fmt: off
@dataclass
class Vector:
    """Position vector"""
    x: float | None = 0
    y: float | None = 0
    z: float | None = 0

    def to_ue_cmd(self, name, cmd):
        """Convert the vector into a command line argument"""
        cmd.append(f"-{name}=(X={self.x},Y={self.y},Z={self.z})")


@dataclass
class Rotation:
    """Rotation vector"""
    pitch   : float | None = 0
    yaw     : float | None = 0
    roll    : float | None = 0

    def to_ue_cmd(self, name, cmd):
        """Convert the rotation into a command line argument"""
        cmd.append(f"-{name}=(Pitch={self.pitch},Yaw={self.yaw},Roll={self.roll})")


@dataclass
class DevArguments:
    """Development arguments"""
    abslog                   : str | None  = None  # Same as LOG= but without a filename length check.
    allusers                 : bool = False  # Add the game for all users when INSTALLGE is specified.
    auto                     : bool = False  # Assume yes on all questions. (for example during compile)
    autocheckoutpackages     : bool = False  # Automatically checkout packages that need to be saved.
    automatedmapbuild        : bool = False  # Perform an automated build of a specified map.
    biascompressionforsize   : bool = False  # Override compression settings with respect to size.
    buildmachine             : bool = False  # Set as build machine. Used for deciding if debug output is enabled.
    bulkimportingsounds      : bool = False  # Use when importing sounds in bulk. (Content Browser specific.)
    check_native_class_sizes : bool = False  # Enable checking of native class sizes. Note: Native classes on console platforms will cause native class size checks to fail even though they are assumed to be correct.
    codermode                : bool = False  # Enables Coder mode.
    compatscale              : bool = False  # Set compatibility settings manually to override PCCompat tool settings.
    conformdir               : bool = False  # Directory to use when conforming packages.
    cookfordemo              : bool = False  # Specify as cooking packages for the demo.
    cookpackages             : bool = False  # Tag to specify cooking packages.
    crashreports             : bool = False  # Always report crashes of the engine.
    d3ddebug                 : bool = False  # Use a d3d debug device.
    devcon                   : bool = False  # Disable secure connections for developers. (Uses unencrypted sockets.)
    dumpfileiostats          : bool = False  # Track and log File IO statistics.
    fixedseed                : bool = False  # Initialize the random number generator with a fixed value, 0.
    fixuptangents            : bool = False  # Fix legacy tangents in distributions automatically.
    forcelogflush            : bool = False  # Force a log flush after each line.
    forcepvrtc               : bool = False  # Force pvrtc texture compression for mobile platform.
    forcesoundrecook         : bool = False  # Force a complete re-cook of all sounds.
    genericbrowser           : bool = False  # Use the Generic Browser.
    installed                : bool = False  # For development purposes, run the game as if installed.
    installfw                : bool = False  # Set whether the handling of the firewall integration should be performed.
    uninstallfw              : bool = False  # Set whether the handling of the firewall integration should be performed.
    installge                : bool = False  # Add the game to the Game Explorer.
    cultureforcooking        : bool = False  # Set language to be used for cooking.
    lightmassdebug           : bool = False  # Launch lightmass manually with -debug and allow lightmass to be executed multiple times.
    lightmassstats           : bool = False  # Force all lightmass agents to report detailed stats to the log.
    log                      : str | None  = None  # When used as a switch (-log), opens a separate window to display the contents of the log in real time. When used as a setting (LOG=filename.log), tells the engine to use the log filename of the string that immediately follows.
    logtimes                 : bool = False  # Print time with log output. (Default, same as setting LogTimes=True in the [LogFiles] section of *Engine.ini.)
    noconform                : bool = False  # Tell the engine not to conform packages as they are compiled.
    nocontentbrowser         : bool = False  # Disable the Content Browser.
    noinnerexception         : bool = False  # Disable the exception handler within native C++.
    noloadstartuppackages    : bool = False  # Force startup packages not to be loaded. You can use this if objects in a startup package must be deleted from within the editor.
    nologtimes               : bool = False  # Do not print time with log output. (Same as setting LogTimes=False in the [LogFiles] section of *Engine.ini.)
    nopause                  : bool = False  # Close the log window automatically on exit.
    nopauseonsuccess         : bool = False  # Close the log window automatically on exit as long as no errors were present.
    norc                     : bool = False  # Disable the remote control. Used for dedicated servers.
    noverifygc               : bool = False  # Do not verify garbage compiler assumptions.
    nowrite                  : bool = False  # Disable output to log.
    seekfreeloading          : bool = False  # Only use cooked data.
    seekfreepackagemap       : bool = False  # Override the package map with the seekfree (cooked) version.
    seekfreeloadingpcconsole : bool = False  # Only use cooked data for PC console mode.
    seekfreeloadingserver    : bool = False  # Only use cooked data for server.
    setthreadnames           : bool = False  # (Xbox only) Force thread names to be set. This can mess up the XDK COM API which is why it must be explicitly set to be performed if desired.
    showmissingloc           : bool = False  # If missing localized text, return error string instead of English text.
    silent                   : bool = False  # Disable output and feedback.
    traceanimusage           : bool = False  # Trace animation usage.
    treatloadwarningsaserrors: bool = False  # Force load warnings to be treated as errors.
    unattended               : bool = False  # Set as unattended. Disable anything requiring feedback from user.
    uninstallge              : bool = False  # Remove the game from the Game Explorer.
    useunpublished           : bool = False  # Force packages in the Unpublished folder to be used.
    vadebug                  : bool = False  # Use the Visual Studio debugger interface.
    verbose                  : bool = False  # Set compiler to use verbose output.
    verifygc                 : bool = False  # Force garbage compiler assumptions to be verified.
    warningsaserrors         : bool = False  # Treat warnings as errors.


@dataclass
class RenderingArguments:
    """Rendering arguments"""
    consolex                : int | None   = None  # Set the horizontal position for console output window.
    consoley                : int | None   = None  # Set the vertical position for console output window.
    winx                    : int | None   = None  # Set the horizontal position of the game window on the screen.
    winy                    : int | None   = None  # Set the vertical position of the game window on the screen.
    resx                    : int | None   = None  # Set horizontal resolution for game window.
    resy                    : int | None   = None  # Set vertical resolution for game window.
    vsync                   : bool | None  = None  # Activate the VSYNC via command line. Pprevents tearing of the image but costs fps and causes input latency.)
    novsync                 : bool | None  = None  # Deactivate the VSYNC via command line.
    benchmark               : bool | None  = None  # Run game at fixed-step in order to process each frame without skipping any frames. This is useful in conjunction with DUMPMOVIE options.
    dumpmovie               : bool | None  = None  # Dump rendered frames to files using current resolution of game.
    exec                    : str | None   = None  # Executes the specified exec file.
    fps                     : int | None   = None  # Set the frames per second for benchmarking.
    fullscreen              : bool | None  = None  # Set game to run in fullscreen mode.
    seconds                 : float | None = None  # Set the maximum tick time.
    windowed                : bool | None  = None  # Set game to run in windowed mode.


@dataclass
class NetworkArguments:
    """Network arguments"""
    lanplay             : bool = False  # Tell the engine to not cap client bandwidth when connecting to servers. Causes double the amount of server updates and can saturate client's bandwidth.
    limitclientticks    : bool = False  # Force throttling of network updates.
    multihome           : bool = False  # Tell the engine to use a multihome address for networking.
    networkprofiler     : bool = False  # Enable network profiler tracking.
    nosteam             : bool = False  # Set steamworks to not be used.
    port                : int | None  = None  # Tell the engine to use a specific port number.
    primarynet          : bool = False  # Affect how the engine handles network binding.


@dataclass
class UserArguments:
    """User Arguments"""
    nohomedir           : bool = False  # Override use of My Documents folder as home directory.
    noforcefeedback     : bool = False  # Disable force feedback in the engine.
    nosound             : bool = False  # Disable any sound output from the engine.
    nosplash            : bool = False  # Disable use of splash image when loading the game.
    notexturestreaming  : bool = False  # Disable texture streaming. Highest quality textures are always loaded.
    onethread           : bool = False  # Run the engine using a single thread instead of multi-threading.
    paths               : str | None  = None  # Set what paths to use for testing wrangled content. Not used for shipping releases.
    preferredprocessor  : bool = False  # Set the thread affinity for a specific processor.
    useallavailablecores: bool = False  # Force the use of all available cores on the target platform.


@dataclass
class ServerArguments:
    """Server arguments"""
    login               : bool = False  # Set username to use when logging in.
    password            : bool = False  # Set password to use when logging in.


@dataclass
class GameStatsArguments:
    """Game stats arguments"""
    nodatabase          : bool = False  # Do not use database, and ignore database connection errors.
    nolivetags          : bool = False  # Skip loading unverified tag changes from SQL database. Only load for current user.


@dataclass
class ConfigArguments:
    """Configuration related arguments."""
    englishcoalesced        : bool = False  # Revert to the default (English) coalesced .ini if the language-localized version cannot be found.
    noautoiniupdate         : bool = False  # Suppress prompts to update .ini files.
    noini                   : bool = False  # Do not update the .ini files.
    regenerateinis          : bool = False  # Forces .ini files to be regenerated.

    defeditorini            : str | None = None  # Set the default editor .ini file to use.
    editorini               : str | None = None  # Set the editor .ini file to use.
    defeditorusersettingsini: str | None = None  # Set the default editor user settings .ini file to use.
    editorusersettingsini   : str | None = None  # Set the editor user settings .ini file to use.
    defcompatini            : str | None = None  # Set the default  compatibility .ini file to use.
    compatini               : str | None = None  # Set the compatibility .ini file to use.
    deflightmassini         : str | None = None  # Set the default  lightmass .ini file to use.
    lightmassini            : str | None = None  # Set the lightmass .ini file to use.
    defengineini            : str | None = None  # Set the default  engine .ini file to use.
    engineini               : str | None = None  # Set the engine .ini file to use.
    defgameini              : str | None = None  # Set the default  game .ini file to use.
    gameini                 : str | None = None  # Set the game .ini file to use.
    definputini             : str | None = None  # Set the default  input .ini file to use.
    inputini                : str | None = None  # Set the input .ini file to use.
    defuiini                : str | None = None  # Set the default  UI .ini file to use.
    uiini                   : str | None = None  # Set the UI .ini file to use.


@dataclass
class DebugArguments:
    """Debugging arguments"""
    bugloc          : Vector | None   = None  # (e.g. BugLoc=(X=1798.8569,Y=475.9513,Z=-8.8500))
    bugrot          : Rotation | None = None  # (e.g. BugRot=(Pitch=-1978,Yaw=-7197,Roll=0))


@dataclass
class MiscArguments:
    """Other arguments"""
    timelimit       : float | None = None  # (e.g. timelimit=[time])
    goalscore       : float | None = None  # (e.g. goalscore=[score])
    numbots         : int | None   = None  # (e.g. numbots=[num])


@dataclass
class CLIArguments:
    """Arguments to make the editor run in command line mode."""
    fullstdoutlogoutput : bool = False  # Force all log to be output to stdout.
    utf8output          : bool = False  # Force all output to be UTF-8.
    nullrhi             : bool = False  # Prevent rendering
    warningsaserror     : bool = False  # Treat warnings as errors.


@dataclass
class Arguments:
    """Unreal Editor arguments"""
    # project: str # Path to the pro
    map     : str | None  = None  # Path to the map
    game    : bool = False        # Launch the game
    server  : bool = False        # Launch the game as a server
    address : str | None = None   # Address to connect to

# fmt: on


class Editor(Command):
    """Runs Editor as is. This command exposes a lot of arguments.

    Notes
    -----

    To make your intention clear, you should implement your own Command with a more defined goal.

    Examples
    --------

    .. code-block:: console

       # Launch the uncooked game (no editor, standalone)
       uecli editor RTSGame -game

       # Launch a listen server (Host other clients, 1 local player)
       uecli editor RTSGame /Game/Maps/MyMap?Listen --game

       # Launch a dedicated server (no local players)
       uecli editor RTSGame /Game/Maps/MyMap --server --game --port 8123

       # Launch a client (1 local player)
       uecli editor RTSGame --address localhost --port 8123 --game

    """

    name: str = "editor"

    @staticmethod
    def examples():
        return [
            "uecli editor RTSGame -game",
            "uecli editor RTSGame /Game/Maps/MyMap --server --game --port 8123",
        ]

    @staticmethod
    def arguments(subparsers):
        """Adds the arguments for this command to the given parser"""
        parser = newparser(subparsers, Editor)

        parser.add_argument(
            "--project",
            type=str,
            metavar="project",
            default=None,
            help="Project name, example: <project>.uproject",
        )
        parser.add_argument(
            "--cli",
            action="store_true",
            default=False,
            help="Enable a group of arguments to make the editor as a command line tool",
        )

        parser.add_argument(
            "--dry",
            action="store_true",
            default=False,
            help="Print the command it will execute without running it",
        )
        parser.add_arguments(Arguments, dest="args")
        parser.add_arguments(DevArguments, dest="dev")
        parser.add_arguments(RenderingArguments, dest="rendering")
        parser.add_arguments(NetworkArguments, dest="network")
        parser.add_arguments(UserArguments, dest="user")
        parser.add_arguments(ServerArguments, dest="server")
        parser.add_arguments(GameStatsArguments, dest="stats")
        parser.add_arguments(DebugArguments, dest="debug")
        parser.add_arguments(MiscArguments, dest="misc")

    @staticmethod
    def execute(args):
        if args.cli:
            args.fullstdoutlogoutput = True
            args.utf8output = True
            args.nullrhi = True
            args.warningsaserror = True
            args.nosplash = True
            args.nosound = True
            args.nopause = True
            args.unattended = True

        project = vars(args).pop("project")
        if project is not None and not project.endswith(".uproject"):
            project = find_project(project)

        cmd = [editor_cmd()] + ([project] if project else []) + command_builder(args)
        print(" ".join(cmd))

        if not args.dry:
            fmt = Formatter()
            return popen_with_format(fmt, cmd)

        return 0


COMMANDS = Editor



import glob
import importlib
import os
import pkgutil
import traceback

try:
    import uetools.plugins

    PLUGINS = True

except ModuleNotFoundError:
    PLUGINS = False

except ImportError:
    PLUGINS = False


def fetch_factories(registry, base_module, base_file_name, function_name="COMMANDS"):
    """Loads all the defined commands"""
    module_path = os.path.dirname(os.path.abspath(base_file_name))

    for module_path in glob.glob(os.path.join(module_path, "[A-Za-z]*.py")):
        module_file = module_path.split(os.sep)[-1]

        if module_file == base_file_name:
            continue

        module_name = module_file.split(".py")[0]

        try:
            module = __import__(".".join([base_module, module_name]), fromlist=[""])
        except ImportError:
            print(traceback.format_exc())
            continue

        if hasattr(module, function_name):
            cmd = getattr(module, function_name)
            registry.insert_commands(cmd)


def discover_plugins(module):
    """Discover uetools plugins"""
    path = module.__path__
    name = module.__name__

    plugins = {}

    for _, name, _ in pkgutil.iter_modules(path, name + "."):
        plugins[name] = importlib.import_module(name)

    return plugins


def discover_from_plugins_commands(registry, module, function_name="COMMANDS"):
    """Imports all commands for the plugins we found"""
    plugins = discover_plugins(module)

    for _, plugin in plugins.items():

        if hasattr(plugin, function_name):
            plugin_commands = getattr(plugin, function_name)
            registry.insert_commands(plugin_commands)


# pylint: disable=too-few-public-methods
class CommandRegistry:
    """Simple class to keep track of all the commands we find"""

    def __init__(self):
        self.found_commands = {}

    def insert_commands(self, cmds):
        """Insert a command into the registry makes sure it is unique"""
        if not isinstance(cmds, list):
            cmds = [cmds]

        for cmd in cmds:
            if cmd.name != cmd.name.strip():
                print(f"Warning: {cmd.name} has white space before or after the name")

            assert (
                cmd.name not in self.found_commands
            ), f"Duplicate command name: {cmd.name}"
            self.found_commands[cmd.name] = cmd


def discover_commands():
    """Discover all the commands we can find (plugins and built-in)"""
    registry = CommandRegistry()
    fetch_factories(registry, "uetools.commands", __file__)

    if PLUGINS:
        discover_from_plugins_commands(registry, uetools.plugins)

    return registry.found_commands


commands = discover_commands()


def find_command(name):
    return commands.get(name)



import logging
import sys
from dataclasses import dataclass

from uetools.core.command import Command, newparser
from uetools.format.base import Formatter
from uetools.format.cooking import CookingFormatter
from uetools.format.tests import TestFormatter

log = logging.getLogger()


profiles = {
    None: Formatter,
    "cooking": CookingFormatter,
    "tests": TestFormatter,
}


@dataclass
class Arguments:
    """Format UnrealEngine log output. It will attempt to align log output to make them more readable.

    Attributes
    ----------
    profile: str
        Formatting profile to use (None, cooking, tests)

    file: str
        File to format, if none it will use stdin

    fail_on_error:
        the program will exit with an error code if errors were found

    col: int
        The size of the category column

    Examples
    --------

    .. code-block:: console

       uecli fmt --profile cooking --file RTSGame.log

       ../UnrealEditor ... | uecli fmt
       [  0][L][LogWindows           ] Failed to load 'aqProf.dll' (GetLastError=126)
       [  0][L][LogWindows           ] File 'aqProf.dll' does not exist
       [  0][L][LogProfilingDebugging] Loading WinPixEventRuntime.dll for PIX profiling (from ../../../Engine/Binaries/ThirdParty/Windows/WinPixEventRuntime/x64).
       [  0][D][LogConfig            ]  Loading HoloLens ini files took 0.02 seconds
       [  0][D][LogConfig            ]  Loading Android ini files took 0.02 seconds
       [  0][D][LogConfig            ]  Loading Unix ini files took 0.03 seconds
       [  0][D][LogConfig            ]  Loading Windows ini files took 0.03 seconds
       [  0][D][LogConfig            ]  Loading TVOS ini files took 0.03 seconds
       [  0][D][LogConfig            ]  Loading Linux ini files took 0.03 seconds
       [  0][D][LogConfig            ]  Loading LinuxArm64 ini files took 0.03 seconds
       [  0][L][LogPluginManager     ] Mounting Engine plugin FastBuildController

    """

    profile: str = None
    file: str = None
    fail_on_error: bool = False
    col: int = 24


class Format(Command):
    """Format UnrealEngine log output. It will attempt to align log output to make them more easily readable."""

    name: str = "format"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, Format)
        parser.add_arguments(Arguments, dest="args")

    def __init__(self, profile=None):
        self.profile = profile

    @staticmethod
    def execute(args):
        args = args.args
        fmt = profiles.get(args.profile, Formatter)(args.col)

        if args.file is not None:
            with open(args.file, encoding="utf-8") as file:
                for line in file:
                    fmt.match_regex(line)

            if args.fail_on_error and len(fmt.bad_logs) > 0:
                return 1
            return 0

        for line in sys.stdin:
            fmt.match_regex(line)

        print("-" * 80)
        print("    Summary")
        print("=" * 80)
        for line in fmt.bad_logs:
            print("  - ", end="")
            Formatter.format(fmt, **line)
        print("=" * 80)

        if args.fail_on_error and len(fmt.bad_logs) > 0:
            return 1

        return 0


logging.basicConfig(level=logging.CRITICAL)

COMMANDS = Format



import logging
import re

import colorama
from colorama import Fore, Style

from uetools.core.conf import load_conf, update_conf

log = logging.getLogger()

UE_LOG_FORMAT = re.compile(
    r"^(\[(?P<datetime>.*)\]\[\s*(?P<frame>\d*)\])?(?P<category>[A-Za-z]*): ((?P<verbosity>[A-Za-z]*):)?(?P<message>.*)"
)

UE_LOG_FORMAT_UTC = re.compile(
    r"^\[(?P<datetime>\d{4}\.\d{2}\.\d{2}-\d{2}\.\d{2}\.\d{2}:\d{3})\]\[\s*(?P<frame>\d*)\](?P<category>[A-Za-z]*): ((?P<verbosity>[A-Za-z]*):)?(?P<message>.*)"
)

UE_STDOUT_FORMAT = re.compile(
    r"^(?P<category>[A-Za-z]*): ((?P<verbosity>[A-Za-z]*):)?(?P<message>.*)"
)

log_verbosity = [
    "Fatal",
    "Error",
    "Warning",
    "Display",
    "Log",
    "Verbose",
    "VeryVerbose",
]

bad_logs = ["Fatal", "Error", "Warning"]

short = dict(
    Fatal="F",
    Error="E",
    Warning="W",
    Display="D",
    Log="L",
    Verbose="V",
    VeryVerbose="T",
)


colors = dict(
    F="red",
    E="red",
    W="yellow",
)


COLORAMA = {
    "red": Fore.RED,
    "blue": Fore.BLUE,
    "green": Fore.GREEN,
    "yellow": Fore.YELLOW,
}

# pylint: disable=unused-argument
def colored(text, color, attrs=None):
    """Returns a colored text"""
    if color is None:
        return text

    return COLORAMA[color.lower()] + text + Style.RESET_ALL


# The logging format is specified in the follorwing function calls
# Logf_InternalImpl(File, Line, Category, Verbosity, Fmt, Args...);
# FPlatformMisc::LowLevelOutputDebugStringf(TEXT("%s%s"),*FOutputDeviceHelper::FormatLogLine(Verbosity, Category, Data, GPrintLogTimes, Time),LINE_TERMINATOR);


class Formatter:
    """Parse an unreal engine output log line and formats it"""

    def __init__(self, col=None) -> None:
        # This is needed for windows
        colorama.init()

        self.col = col
        self.longest_category = load_conf().get("longest_category", 0)
        self.longest_name = load_conf().get("longest_name", "")
        self.regex = UE_LOG_FORMAT
        self.print_non_matching = False
        self.bad_logs = []
        self.ignore = set()
        self.only = set()

        if self.col is not None:
            self.longest_category = self.col

    def _meta(self, category):
        if self.col is not None:
            return

        self.longest_category = max(self.longest_category, len(category))
        self.longest_name = (
            category if len(category) > len(self.longest_name) else self.longest_name
        )

    def summary(self):
        """Print a summary of warnings and errors that got parsed during the formatting process"""
        print("-" * 80)
        print("    Summary")
        print("=" * 80)
        for line in self.bad_logs:
            print("  - ", end="")
            Formatter.format(self, **line)
        print("=" * 80)

    def __del__(self):
        update_conf(longest_category=self.longest_category)

    def match_regex(self, line):
        """Parse a log line using regex"""
        result = self.regex.search(line)

        if result:
            data = result.groupdict()

            if data["verbosity"] is None:
                data["verbosity"] = "Log"

            if data["verbosity"] not in log_verbosity:
                msg = data["verbosity"]
                data["verbosity"] = "Log"
                data["message"] = f"{msg}: " + data["message"]

            # Kepp track of bad logs and show a summary at the end
            if data["verbosity"] in bad_logs:
                self.bad_logs.append(data)

            self.format(**data)
        else:
            if self.print_non_matching:
                print(line, end="")
            else:
                log.debug("    Line did not match anything")
                log.debug("        - `%s`", line)

    # pylint: disable=too-many-arguments
    def format(
        self, datetime=None, frame=None, category=None, verbosity=None, message=None
    ):
        """Creates column for each element of the log and print the message last.

        Parameters
        ----------

        datetime:
            Can be, UTC, local, seconds since start or time code (Frametime converted)

        frame:
            GFrameCounter % 1000

        category:
            log category

        verbosity:
            log verbosity

        message:
            log message

        """
        if category in self.ignore:
            return

        if len(self.only) > 0 and category not in self.only:
            return

        self._meta(category)
        verb = short.get(verbosity, " ")

        if verb == " ":
            log.debug("%s", verbosity)

        category = f"{{:<{self.longest_category}}}".format(category)

        color = colors.get(verb)

        if frame is None:
            frame = 0

        print(f"[{int(frame):3d}][{verb}][{category}] {colored(message, color=color)}")



from dataclasses import dataclass

from uetools.core.command import Command, newparser
from uetools.core.conf import editor, find_project
from uetools.core.run import run


# This is not used technically, but we keep it for consistency with the other commands
# also help with the doc generation as this object appears on top
@dataclass
class Arguments:
    """Open the editor for a given project

    Attributes
    ----------
    name: str
        Name of the the target to build (UnrealPak, RTSGame, RTSGameEditor, etc...)

    Examples
    --------

    .. code-block:: console

       uecli open RTSGameEditor

    """

    name: str  # Name of the the target to build (UnrealPak, RTSGame, RTSGameEditor, etc...)


class Open(Command):
    """Open the editor for a given project"""

    name: str = "open"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, Open)
        # this makes it ugly
        # editor.add_arguments(Arguments, dest="open")

        parser.add_argument("name", type=str, help="Name of the the project to open")

    @staticmethod
    def execute(args):
        project = find_project(args.name)

        return run([editor(), project], check=True).returncode


COMMANDS = Open



import subprocess


def git_status(repo):
    """Return git status of a given repo"""
    return subprocess.check_output(["git", "status"], cwd=repo, text=True)


def git_diff(repo):
    """Return git diff of a given repo"""
    return subprocess.check_output(["git", "diff"], cwd=repo, text=True)



import json
import os
import shutil
import tempfile
from dataclasses import dataclass
from typing import Optional

import pkg_resources
from cookiecutter.main import cookiecutter

from uetools.core.command import Command, command_builder, newparser
from uetools.core.conf import (
    engine_folder,
    find_project,
    get_build_platforms,
    get_version_tag,
    guess_platform,
    retrieve_exact_engine_version,
    uat,
)
from uetools.core.run import popen_with_format
from uetools.format.cooking import CookingFormatter


# fmt: off
@dataclass
class Arguments:
    """Builds and cook a plugin"""

    # project             : str # Name of the project
    # plugin              : str # Path to the plugin (relative to project folder)
    # platforms           : List[str] = field(default_factory=list)
    # output              : str  = None  # Packaged plugin destination
    StrictIncludes      : bool = False  # Disables precompiled headers and unity build in order to check all source files have self-contained headers.
    NoHostPlatform      : bool = False  # Prevent compiling for the editor platform on the host
    verbose             : bool = False  # Enables verbose logging
    veryverbose         : bool = False  # Enables very verbose logging
    submit              : bool = False  # Allows UAT command to submit changes
    nosubmit            : bool = False  # Prevents any submit attempts
    np4                 : bool = False  # Disables Perforce functionality
    nonp4               : bool = False  # Enables Perforce functionality
    NoKill              : bool = False  # Does not kill any spawned processes on exit
    Compile             : bool = False  # Force all script modules to be compiled
    NoCompile           : bool = False  # Do not attempt to compile any script modules
    IgnoreBuildRecords  : bool = False  # Ignore build records (Intermediate/ScriptModule/ProjectName.json)
    UseLocalBuildStorage: bool = False  # Allows you to use local storage for your root build storage dir
    WaitForDebugger     : bool = False  # Waits for a debugger to be attached, and breaks once debugger successfully attached.
    Unversioned         : bool = False  # Do not embed the current engine version into the descriptor
    Rocket              : bool = True   # Undocumented argument
    EngineDir           : Optional[str] = None  # Engine directory
    CreateSubFolder     : bool = False  # Create a subfolder for the plugin
    NoPCH               : bool = False  # No Precompiled Header
    NoSharedPCH         : bool = False  # No Shared Precompiled Header
    DisableUnity        : bool = False  # Disable Unity Build
    NoDeleteHostProject : bool = False  # Do not delete host project (which was created to prepare the plugin)
    # this does not exist when calling BuildPlugin
    # UbtArgs             : Optional[str]     = None  # extra options to pass to ubt
# fmt: on


class PackagePlugin(Command):
    """Builds and cook a plugin

    Examples
    ---------

    Path to a plugin that is not inside a project

    .. code-block:: python

       uetools plugin-cook /Projects/Gamekit/Gamekit.uplugin --output /Built/Gamekit/ --platforms Win64


    Path to a plugin inside a project

    .. code-block:: python

       uetools plugin-cook --project MyHostProject Plugins/Gamekit/Gamekit.uplugin --output /Built/Gamekit/ --platforms Win64

    """

    name: str = "plugin-cook"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, PackagePlugin)
        parser.add_argument("plugin", type=str, help="Path to uplugin file")
        parser.add_argument(
            "--project", type=str, help="Project name containing the plugin"
        )
        parser.add_argument(
            "--platforms",
            type=str,
            nargs="+",
            choices=get_build_platforms(),
            default=[guess_platform()],
            help="list of platforms to build for",
        )
        parser.add_argument(
            "--output", type=str, help="path to build the packaged plugin"
        )

        parser.add_arguments(Arguments, dest="pkg")

    @staticmethod
    def execute(args):
        project = args.project
        plugin = args.plugin
        platforms = "+".join(args.platforms)

        if project is not None and not os.path.isabs(plugin):
            project = find_project(args.project)
            folder = os.path.dirname(project)
            plugin_path = os.path.join(folder, plugin)

        else:
            plugin_path = plugin

        if not os.path.exists(plugin_path):
            raise RuntimeError(f"Could not find {plugin_path}")

        plugin_path = os.path.abspath(plugin_path)

        cmdargs = (
            [
                uat(),
                "BuildPlugin",
                f"-Plugin={plugin_path}",
                f"-Package={args.output}",
                f"-TargetPlatforms={platforms}",
            ]
            + command_builder(args.pkg)
            + ["-nocompileuat"]
        )

        print(" ".join(cmdargs))

        fmt = CookingFormatter(24)
        fmt.print_non_matching = True
        return popen_with_format(fmt, cmdargs)


class NewPlugin(Command):
    """Create a new plugin from a template"""

    name: str = "plugin-new"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, NewPlugin)
        parser.add_argument(
            "project", type=str, help="Project in which the plugin will live"
        )
        parser.add_argument("plugin", type=str, help="Name of the plugin")

    @staticmethod
    def execute(args):
        project = find_project(args.project)
        project_dir = os.path.dirname(project)

        template = pkg_resources.resource_filename(
            __name__, "../templates/PluginTemplate/cookiecutter.json"
        )
        assert os.path.exists(template)
        template = os.path.dirname(template)
        assert os.path.exists(template)

        configfile = tempfile.NamedTemporaryFile("w", suffix=".json", delete=False)
        json.dump({"default_context": {"plugin_name": args.plugin}}, configfile)
        configfile.flush()

        plugin_dir = os.path.join(project_dir, "Plugins")
        assert os.path.exists(plugin_dir)

        kwargs = dict(
            no_input=True,
            config_file=configfile.name,
            overwrite_if_exists=True,
            output_dir=plugin_dir,
        )

        cookiecutter(
            template,
            **kwargs,
        )

        # Windows have permission issues on reading a temporary files
        configfile.close()
        os.remove(configfile.name)


class FinalizePlugin(Command):
    """Finalize Plugin for redistribution

    * Set the engine version inside the <plugin>.uplugin
    * Set installed to false inside <plugin>.uplugin
    * Check MarketplaceURL
    * Set VersionName
    * Copy some Config folder

    """

    name: str = "plugin-finalize"

    @dataclass
    class Arguments:
        plugin: str  # plugin name
        output: str  # output

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, FinalizePlugin)
        parser.add_arguments(FinalizePlugin.Arguments, dest="args")

    @staticmethod
    def execute(args):
        args = args.args
        base_url = "com.epicgames.launcher://ue/marketplace/product/"
        plugin_dir = os.path.dirname(os.path.abspath(args.plugin))

        plugin_version = get_version_tag(plugin_dir).replace("v", "").split("-")[0]

        engine_version = retrieve_exact_engine_version(engine_folder())

        # Configure the plugin descriptor
        # -------------------------------
        with open(args.output) as f:
            uplugin = json.load(f)

        version_old = uplugin["VersionName"]
        installed_old = uplugin["Installed"]
        engine_old = uplugin["EngineVersion"]

        uplugin["VersionName"] = plugin_version
        uplugin["Installed"] = False
        uplugin["EngineVersion"] = engine_version

        print("Plugin Version:", version_old, " => ", plugin_version)
        print("     Installed:", installed_old, " => ", False)
        print(" EngineVersion:", engine_old, " => ", engine_version)

        assert (
            len(uplugin["MarketplaceURL"][len(base_url) :]) > 0
        ), "MarketPlace URL missing"

        with open(args.output, "w") as f:
            json.dump(uplugin, f)

        # Copy files
        # ----------
        config_folder = os.path.join(plugin_dir, "Config")
        output_folder = os.path.dirname(args.output)

        shutil.copytree(
            config_folder, os.path.join(output_folder, "Config"), dirs_exist_ok=True
        )


COMMANDS = [PackagePlugin, NewPlugin, FinalizePlugin]



import os

from cookiecutter.main import cookiecutter

from uetools.core.command import Command, newparser
from uetools.core.conf import find_project

COOKIECUTTER = "https://github.com/kiwi-lang/UEDocs"


# pylint: disable=too-few-public-methods
class Arguments:
    """Add a Sphinx + Doxygen documentation to a project

    Attributes
    ----------
    name: str
        Name of the project

    """

    name: str


class Docs(Command):
    """Add a docs folder to your project"""

    name: str = "docs"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, Docs)
        parser.add_argument("project", type=str, help="name of your project")
        parser.add_argument(
            "--no-input",
            action="store_true",
            default=False,
            help="Do not show user prompts",
        )
        parser.add_argument(
            "--config",
            type=str,
            default=None,
            help="Configuration file used to initialize the project (json)",
        )

    @staticmethod
    def execute(args):
        uproject = find_project(args.project)
        folder = os.path.dirname(uproject)

        os.chdir(folder)
        cookiecutter(
            COOKIECUTTER,
            no_input=args.no_input,
            config_file=args.config,
            overwrite_if_exists=True,
        )

        return 0


COMMANDS = Docs



import os
from argparse import Namespace
from dataclasses import asdict, dataclass
from typing import Optional

from simple_parsing import choice

from uetools.commands.build import Build
from uetools.core.command import Command, command_builder, newparser
from uetools.core.conf import (
    build_platform_from_editor,
    editor_cmd,
    find_project,
    get_build_modes,
    get_build_platforms,
    get_editor_platforms,
    guess_editor_platform,
    guess_platform,
    uat,
)
from uetools.core.run import popen_with_format
from uetools.format.cooking import CookingFormatter


@dataclass
class Arguments:
    """Cook your main game

    Attributes
    ----------
    name: str
        Name of the project to cook

    output: str
        Cooking result output path, defaults to ``$PROJECT_NAME/Saved/StagedBuilds/``

    build: str
        Build mode, if set a build command will be issued before cooking

    platform: str
        Used if build mode is set.

    compressed: bool
        Compress the cooked data, defaults to ``True``

    cookall: bool
        Cook all assets, defaults to ``False``

    unversioned: bool
        Don't version the cooked data, defaults to ``False``

    WarningsAsErrors: bool
        Treat warnings as errors, defaults to ``False``

    Notes
    -----

    UAT builds the game before executing the cook command;  You should too.

    Examples
    --------

    .. code-block:: console

       # Only cooks the game
       uecli cook RTSGame --platform Windows

       # Build the project for development before Cooking
       uecli cook RTSGame --platform Windows --build Development

    """

    project: str
    output: Optional[str] = None
    build: Optional[str] = choice(*get_build_modes(), default=None)
    platform: Optional[str] = choice(
        *get_editor_platforms(), default=guess_editor_platform()
    )
    compressed: bool = True
    cookall: bool = True
    unversioned: bool = True
    WarningsAsErrors: bool = True


class CookGame(Command):
    """Cook your main game"""

    name: str = "cook"

    @staticmethod
    def arguments(subparsers):
        """Add arguments to the parser"""
        parser = newparser(subparsers, CookGame)
        parser.add_arguments(Arguments, dest="cook")

    @staticmethod
    def execute(args):
        """Execute the cook command using the editor"""
        args = Namespace(**asdict(args.cook))

        name = vars(args).pop("project")
        platform = vars(args).pop("platform")
        build = vars(args).pop("build")

        if build:
            build_args = Namespace()
            build_args.target = name
            build_args.platform = build_platform_from_editor(platform)
            build_args.mode = build
            build_args.profile = "update-project"
            Build.execute_profile(Namespace(build=build_args))

        uproject = find_project(name)
        folder = os.path.dirname(uproject)

        if args.output is None:
            args.output = os.path.join(folder, "Saved", "StagedBuilds")

        cli_cmd = [
            "-CrashForUAT",
            "-unattended",
            "-NoLogTimes",
            "-UTF8Output",
            "-FullStdOutLogOutput",
        ]

        options = command_builder(args)

        cmd = (
            [
                editor_cmd(),
                uproject,
                "-run=cook",
                f"-targetplatform={platform}",
                f"-abslog={folder}/Saved/Automation/Cooking.txt",
            ]
            + cli_cmd
            + options
        )

        print(" ".join(cmd), flush=True)

        fmt = CookingFormatter(24)
        fmt.print_non_matching = True
        returncode = popen_with_format(fmt, cmd)
        fmt.summary()

        print(f"Subprocess terminated with (rc: {returncode})")

        return returncode


@dataclass
class UATArguments:
    """Cook arguments for UAT"""

    project: str
    unattended: bool = True
    utf8output: bool = True
    platform: str = choice(*get_build_platforms(), default=guess_platform())
    clientconfig: str = choice(*get_build_modes(), default="Development")
    serverconfig: str = choice(*get_build_modes(), default="Development")
    noP4: bool = True
    nodebuginfo: bool = True
    allmaps: bool = True
    cook: bool = True
    build: bool = True
    stage: bool = True
    prereqs: bool = True
    pak: bool = True
    archive: bool = True
    stagingdirectory: Optional[str] = None
    archivedirectory: Optional[str] = None
    WarningsAsErrors: bool = True


class CookGameUAT(Command):
    """Cook your main game using UAT"""

    name: str = "uat-cook"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, CookGameUAT)
        parser.add_arguments(UATArguments, dest="cook")

    @staticmethod
    def execute(args):
        args = args.cook

        args.project = find_project(args.project)

        uat_args = command_builder(asdict(args))
        cmd = [uat()] + ["BuildCookRun"] + uat_args + ["-nocompileuat"]

        print(" ".join(cmd))

        fmt = CookingFormatter(24)
        fmt.print_non_matching = True
        returncode = popen_with_format(fmt, cmd, shell=False)
        fmt.summary()

        return returncode


COMMANDS = [
    CookGame,
    CookGameUAT,
]



import os
from pathlib import Path

from setuptools import setup

with open("uetools/core/__init__.py") as file:
    for line in file.readlines():
        if "version" in line:
            version = line.split("=")[1].strip().replace('"', "")
            break

assert (
    os.path.exists(os.path.join("uetools", "__init__.py")) is False
), "uetools is a namespace not a module"


setup_kwargs = dict(
    name="uetools",
    version=version,
    description="Tools for Unreal Engine",
    long_description=(Path(__file__).parent / "README.rst").read_text(),
    author="Pierre Delaunay",
    author_email="pierre@delaunay.io",
    license="BSD-3-Clause",
    url="https://uetools.readthedocs.io",
    packages=[
        "uetools.core",
        "uetools.commands",
        "uetools.format",
        "uetools.plugins.ml",
    ],
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],
    zip_safe=True,
    setup_requires=["setuptools"],
    install_requires=[
        "appdirs",
        "colorama",
        "cookiecutter",
        "simple-parsing",
        "GitPython",
    ],
    python_requires=">=3.7.*",
    namespace_packages=["uetools", "uetools.plugins"],
    entry_points={
        "console_scripts": [
            "uecli = uetools.core.cli:main",
        ],
        "UECLICommands": [
            "ml = uetools.plugins.ml:ML",
            "cpp = uetools.plugins.cpp:CPP",
        ],
    },
    package_data={"uetools": ["uetools/templates/.*"]},
    include_package_data=True,
)

if __name__ == "__main__":
    setup(**setup_kwargs)



import os
from dataclasses import dataclass
from typing import Optional

from simple_parsing import choice

from uetools.core.command import Command, newparser
from uetools.core.conf import (
    engine_folder,
    find_project,
    get_build_modes,
    get_build_platforms,
    guess_platform,
    ubt,
)
from uetools.core.run import run

project_uht = [
    "UnrealHeaderTool",
    "{PLATFORM}",
    "{MODE}",
    "-Project={UPROJECT}",
    "{UPROJECT}",
    "-NoUBTMakefiles",
    "-Manifest={ENGINE_FOLDER}/Intermediate/Build/Manifest.xml",
    "-NoHotReload",
    '-abslog="{ENGINE_FOLDER}/Programs/AutomationTool/Saved/Logs/UBT-UnrealHeaderTool-{PLATFORM}-{MODE}.txt"',
]

engine_uht = [
    "UnrealHeaderTool",
    "{PLATFORM}",
    "{MODE}",
    "-NoUBTMakefiles"
    "-Manifest={ENGINE_FOLDER}/Intermediate/Build/Manifest.xml"
    "-NoHotReload",
    '-abslog="{ENGINE_FOLDER}/Programs/AutomationTool/Saved/Logs/UBT-UnrealHeaderTool-{PLATFORM}-{MODE}.txt"',
]

project_editor = [
    "{PROJECT_NAME}Editor",
    "{PLATFORM}",
    "{MODE}",
    "-Project={UPROJECT}",
    "{UPROJECT}",
    "-NoUBTMakefiles",
    "-Manifest={ENGINE_FOLDER}/Intermediate/Build/Manifest.xml",
    "-NoHotReload",
    '-abslog="{ENGINE_FOLDER}/Programs/AutomationTool/Saved/Logs/UBT-{PROJECT_NAME}Editor-{PLATFORM}-{MODE}.txt"',
]

shader_compile_worker = [
    "ShaderCompileWorker",
    "{PLATFORM}",
    "{MODE}",
    "-NoUBTMakefiles",
    "-Manifest={ENGINE_FOLDER}/Intermediate/Build/Manifest.xml",
    "-NoHotReload",
    '-abslog="{ENGINE_FOLDER}/Programs/AutomationTool/Saved/Logs/UBT-ShaderCompileWorker-{PLATFORM}-{MODE}.txt"',
]

unrealpak = [
    "UnrealPak",
    "{PLATFORM}",
    "{MODE}",
    "-Project={UPROJECT}",
    "{UPROJECT}",
    "-NoUBTMakefiles",
    "-Manifest={ENGINE_FOLDER}/Intermediate/Build/Manifest.xml",
    "-NoHotReload",
    '-abslog="{ENGINE_FOLDER}/Programs/AutomationTool/Saved/Logs/UBT-UnrealPak-{PLATFORM}-{MODE}.txt"',
]

project = [
    "{PROJECT_NAME}",
    "{PLATFORM}",
    "{MODE}",
    "-Project={UPROJECT}",
    "{UPROJECT}",
    "-NoUBTMakefiles",
    "-Manifest={ENGINE_FOLDER}/Intermediate/Build/Manifest.xml",
    "-NoHotReload",
    '-abslog="{ENGINE_FOLDER}/Programs/AutomationTool/Saved/Logs/UBT-{PROJECT_NAME}-{PLATFORM}-{MODE}.txt"',
]

# this is windows only
bootstrap = [
    "BootstrapPackagedGame",
    "{PLATFORM}",
    "Shipping",
    "-NoUBTMakefiles",
    "-Manifest={ENGINE_FOLDER}/Intermediate/Build/Manifest.xml",
    "-NoHotReload",
    '-abslog="{ENGINE_FOLDER}/Programs/AutomationTool/Saved/Logs/UBT-BootstrapPackagedGame-{PLATFORM}-Shipping.txt"',
]


# This is the build commands issued by UAT when using BuildCookRun.
build_update_project = [
    project_uht,
    engine_uht,
    project_editor,
    shader_compile_worker,
    unrealpak,
    project,
    bootstrap,
]


short_update = [
    project_editor,
    shader_compile_worker,
]


profiles = {"update-project": build_update_project, "short-update": short_update}


def replace_variables(command, variables):
    """Replace variables in a command"""
    cmd = []
    for arg in command:
        for key, value in variables.items():
            arg = arg.replace(f"{{{key}}}", value)

        cmd.append(arg)
    return cmd


@dataclass
class Arguments:
    """Execute UnrealBuildTool for a specified target

    Attributes
    ----------
    target: str
        Name of the the target to build (UnrealPak, RTSGame, RTSGameEditor, etc...)

    platform: str
        Platform to build for, defaults to current platform (Win64, Linux, etc..)

    mode: str
        Build mode (Tests, Debug, Development, Shipping)

    profile: str
        Build multiple targets using a configuration

    Examples
    --------

    .. code-block:: console

       # Builds the Editor for the `RTSGame` project
       # Only run the build editor target
       uecli build RTSGameEditor

       # Runs all the command UAT would execute to compile the project (i.e compiles ShaderCompileWorker and others)
       uecli build RTSGame --profile update-project

    Notes
    -----

    While this works (as of 2022-08-21), you should probably rely epics generated commands instead.
    This was done as an exercice to learn about Unreal internals and might not get updated too often.

    """

    target: str
    platform: str = choice(*get_build_platforms(), default=guess_platform())
    mode: str = choice(*get_build_modes(), default="Development")
    profile: Optional[str] = None


class Build(Command):
    """Execute UnrealBuildTool for a specified target"""

    name: str = "build"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, Build)
        parser.add_arguments(Arguments, dest="build")

    @staticmethod
    def execute_profile(args):
        """Execute a command profile"""
        args = args.build

        commands = profiles.get(args.profile)

        uproject = find_project(args.target)
        assert uproject is not None, f"Project {args.target} does not exist"

        name = args.target
        if name.endswith("Editor"):
            name = name[0:-6]

        variables = {
            "PLATFORM": args.platform,
            "MODE": args.mode,
            "UPROJECT": uproject,
            "PROJECT_NAME": name,
            "ENGINE_FOLDER": engine_folder(),
        }

        for command in commands:
            # Temporary fix `BootstrapPackagedGame` is Windows only
            if "BootstrapPackagedGame" in command and args.platform != "Windows":
                continue

            cmd = replace_variables(command, variables)

            cmd = [ubt()] + cmd
            print(" ".join(cmd), flush=True)
            run(cmd, check=True)

        return 0

    @staticmethod
    def execute(args):
        """Execute the UAT build tool on the target"""
        target = args.build.target
        platform = args.build.platform
        mode = args.build.mode

        if args.build.profile:
            return Build.execute_profile(args)

        assert target is not None, "Target name is required"
        engine_path = engine_folder()

        logfolder = os.path.join(engine_path, "Programs/AutomationTool/Saved.Logs")
        logfile = f"UBT-{target}-{platform}-{mode}.txt"
        logfile = os.path.join(logfolder, logfile)
        # "{ENGINE_FOLDER}/Programs/AutomationTool/Saved/Logs/UBT-UnrealHeaderTool-Win64-Development.txt"

        cmd = [
            target,
            platform,
            mode,
        ]

        # Check if the target is a project
        uproject = find_project(target)

        if uproject is not None:
            cmd += [f"-Project={uproject}", uproject]

        cmd.append("-NoUBTMakefiles")
        cmd.append("-NoHotReload")
        cmd.append(f"-Manifest={engine_path}/Intermediate/Build/Manifest.xml")
        cmd.append(f"-abslog={logfile}")

        # Tools
        #    RequiredTools UnrealFrontend UnrealEditor UnrealInsights
        #
        # I have an issue forwarding arguments to linux
        # The make file simply does:
        #     $(BUILD) UnrealHeaderTool Linux Development  -project="$(GAMEPROJECTFILE)" $(ARGS)
        # with:
        #    bash "$(UNREALROOTPATH)/Engine/Build/BatchFiles/Linux/Build.sh"
        #
        # The script then does:
        #   dotnet Engine/Binaries/DotNET/UnrealBuildTool/UnrealBuildTool.dll "$@"
        #
        #
        cmd = [ubt()] + cmd
        print(" ".join(cmd), flush=True)

        return run(cmd, check=True).returncode


COMMANDS = Build



import subprocess

# This is a bit of future proofing in case I start to need to wrap it
run = subprocess.run


def popen_with_format(fmt, args, shell=False):
    """Execute a command with the given formatter."""
    with subprocess.Popen(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        # This is needed because without lines might not be recognized as such
        text=True,
        shell=shell,
    ) as process:
        try:
            while process.poll() is None:
                # sys.stdout.flush()

                line = process.stdout.readline()

                if len(line) > 0:
                    fmt.match_regex(line)

            return process.poll()
        except KeyboardInterrupt:
            print("Stopping due to user interrupt")
            process.kill()

        return -1



from dataclasses import dataclass
from typing import Optional

from uetools.core.command import Command, newparser
from uetools.core.conf import editor, find_project
from uetools.core.run import popen_with_format
from uetools.format.base import Formatter


# This is not used technically, but we keep it for consistency with the other commands
# also help with the doc generation as this object appears on top
@dataclass
class Arguments:
    """Open the editor for a given project

    Attributes
    ----------
    project: str
        Name of the the target to build (UnrealPak, RTSGame, RTSGameEditor, etc...)

    address: str
        Address of the server to connect to, if None launch in standalone

    Examples
    --------

    .. code-block:: console

       uecli client RTSGameEditor

       uecli client RTSGameEditor --address localhost --port 8123

    """

    # name: str
    address: Optional[str] = None
    port: int = 8123  # Server port


class Client(Command):
    """Launch the editor as a client to an already running server"""

    name: str = "client"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, Client)
        parser.add_argument("project", type=str, help="Name of the the project to open")
        parser.add_arguments(Arguments, dest="client")
        parser.add_argument(
            "--dry",
            action="store_true",
            default=False,
            help="Print the command it will execute without running it",
        )

    @staticmethod
    def execute(args):
        project = find_project(args.project)

        cmd = [editor(), project]

        if args.client.address:
            cmd.append(args.client.address)
            cmd.append(f"-port={args.client.port}")

        cmd.append("-game")
        cmd.append("-FullStdOutLogOutput")
        print(" ".join(cmd))

        if not args.dry:
            fmt = Formatter()
            return popen_with_format(fmt, cmd)

        return 0


COMMANDS = Client



import json
from dataclasses import dataclass

from uetools.core.command import Command, newparser
from uetools.core.conf import find_project


@dataclass
class Arguments:
    """Disable a plugin

    Attributes
    ----------
    name: str
        Name of the project to modify

    plugin: str
        Name of the plugin to disable

    Examples
    --------

    .. code-block:: console

       # Disable the plugin
       uecli disable RTSGame RTSGamePlugin

    """

    name: str
    plugin: str


class Disable(Command):
    """Disable unused plugin that are loading by default"""

    name: str = "disable"

    @staticmethod
    def arguments(subparsers):
        parser = newparser(subparsers, Disable)
        parser.add_argument("name", type=str, help="project's name")
        parser.add_argument("plugin", type=str, help="Plugin's name")

    @staticmethod
    def execute(args):
        name = args.name
        uproject = find_project(args.name)

        with open(uproject, encoding="utf-8") as project_file:
            project_conf = json.load(project_file)

        plugins = project_conf.get("Plugins")
        plugin_dict = {}

        for plugin in plugins:
            name = plugin["Name"]
            data = plugin

            plugin_dict[name] = data

        if args.plugin in plugin_dict:
            plugin_dict[args.plugin]["Enabled"] = False
        else:
            plugins.append(dict(Name=args.plugin, Enabled=False))

        with open(uproject, "w", encoding="utf-8") as project_file:
            json.dump(project_conf, project_file)

        return 0


COMMANDS = Disable



from __future__ import annotations

import os
from argparse import Namespace
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import asdict, is_dataclass


def newparser(subparsers, commandcls: Command):
    """Add a subparser to the parser for the command"""
    # The help text is not showing :/
    return subparsers.add_parser(commandcls.name, help=commandcls.help())


@contextmanager
def chdir(root):
    """change directory and revert back to previous directory"""
    old = os.getcwd()
    os.chdir(root)

    yield
    os.chdir(old)


class Command:
    """Base class for all commands"""

    name: str

    @classmethod
    def help(cls) -> str:
        """Return the help text for the command"""
        return cls.__doc__

    @staticmethod
    def arguments(subparsers):
        """Define the arguments of this command"""
        raise NotImplementedError()

    @staticmethod
    def execute(args) -> int:
        """Execute the command"""
        raise NotImplementedError()

    @staticmethod
    def examples() -> list[str]:
        """returns a list of examples"""
        return []


def command_builder(args: dict | Namespace, ignore=None) -> list[str]:
    """Convert a namespace of arguments into a list of command line arguments for unreal engine.
    Supports dataclasses (even nested) and custom command generation through the ``to_ue_cmd`` method.

    Examples
    --------
    >>> from dataclasses import dataclass

    >>> command_builder(dict(log=True, map='/Game/Map/TopDown'))
    ['-log', '-map=/Game/Map/TopDown']

    >>> @dataclass
    ... class Arguments:
    ...     flag       : bool = False
    ...     goalscore  : Optional[float] = None
    ...     something  : Optional[str] = None

    >>> command_builder(dict(vector=Arguments(flag=True, goalscore=2, something=None)))
    ['-flag', '-goalscore=2']

    >>> command_builder(dict(vector=Arguments(flag=False, goalscore=2)))
    ['-goalscore=2']


    >>> @dataclass
    ... class Vector:
    ...     x: Optional[float] = 0
    ...     y: Optional[float] = 0
    ...     z: Optional[float] = 0
    ...     def to_ue_cmd(self, name, cmd):
    ...         cmd.append(f"-{name}=(X={self.x},Y={self.y},Z={self.z})")

    >>> command_builder(dict(vector=Vector(x=1, y=2, z=3)))
    ['-vector=(X=1,Y=2,Z=3)']

    >>> command_builder(Namespace(vector=Vector(x=1, y=2, z=3)))
    ['-vector=(X=1,Y=2,Z=3)']

    """
    if ignore is None:
        ignore = set()

    args = deepcopy(args)

    if isinstance(args, Namespace):
        args = vars(args)

    if not isinstance(args, dict):
        args = asdict(args)

    # Note: we do not NEED to pop them, UE ignore unknown arguments
    if isinstance(args, dict):
        args.pop("command", None)
        args.pop("cli", None)
        args.pop("dry", None)

    cmd = []

    _command_builder(cmd, args, ignore)

    return cmd


def _command_builder(cmd, args, ignore):
    for k, v in args.items():
        if v is None:
            continue

        if k in ignore:
            continue

        if isinstance(v, bool):
            if v is not None and v is True:
                cmd.append(f"-{k}")

        elif isinstance(v, (str, int)):
            cmd.append(f"-{k}={v}")

        elif hasattr(v, "to_ue_cmd"):
            v.to_ue_cmd(k, cmd)

        elif is_dataclass(v):
            _command_builder(cmd, asdict(v), ignore)
